# coding: utf-8

# docs
from __future__ import annotations

from copy import deepcopy
from math import ceil, floor
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

import cv2
import numpy as np

from supervisely.geometry import validation
from supervisely.geometry.constants import (
    CLASS_ID,
    CREATED_AT,
    EXTERIOR,
    ID,
    INTERIOR,
    LABELER_LOGIN,
    POINTS,
    UPDATED_AT,
)
from supervisely.geometry.geometry import Geometry
from supervisely.geometry.point_location import PointLocation, points_to_row_col_list

if TYPE_CHECKING:
    from supervisely.geometry.image_rotator import ImageRotator


class Rectangle(Geometry):
    """Axis-aligned 2D bounding box (top, left, bottom, right). Immutable."""

    @staticmethod
    def geometry_name():
        """
        Returns the name of the geometry.

        :returns: name of the geometry
        :rtype: str
        """
        return "rectangle"

    def __init__(
        self,
        top: int,
        left: int,
        bottom: int,
        right: int,
        sly_id: Optional[int] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[int] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ):
        """
        Axis-aligned 2D bounding box.

        :param top: Top edge (min row).
        :type top: int or float
        :param left: Left edge (min col).
        :type left: int or float
        :param bottom: Bottom edge (max row).
        :type bottom: int or float
        :param right: Right edge (max col).
        :type right: int or float
        :param sly_id: Server-side rectangle ID.
        :type sly_id: int, optional
        :param class_id: ObjClass ID.
        :type class_id: int, optional
        :param labeler_login: Login of user who created the rectangle.
        :type labeler_login: str, optional
        :param updated_at: Last modification timestamp (ISO format).
        :type updated_at: str, optional
        :param created_at: Creation timestamp (ISO format).
        :type created_at: str, optional
        :raises ValueError: If top > bottom or left > right.

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                rect = sly.Rectangle(100, 100, 700, 900)
        """
        if top > bottom:
            raise ValueError(
                'Rectangle "top" argument must have less or equal value then "bottom"!'
            )

        if left > right:
            raise ValueError(
                'Rectangle "left" argument must have less or equal value then "right"!'
            )

        super().__init__(
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

        self._points = [
            PointLocation(row=top, col=left),
            PointLocation(row=bottom, col=right),
        ]

    def to_json(self) -> Dict:
        """
        Convert the Rectangle to a json dict. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :returns: Rectangle in json format as a dict.
        :rtype: dict

        :Usage Example:

            .. code-block:: python

                figure_json = figure.to_json()
                print(figure_json)
                # Output: {
                #    "points": {
                #        "exterior": [
                #            [100, 100],
                #            [900, 700]
                #        ],
                #        "interior": []
                #    }
                # }
        """
        packed_obj = {
            POINTS: {
                EXTERIOR: points_to_row_col_list(self._points, flip_row_col_order=True),
                INTERIOR: [],
            }
        }
        self._add_creation_info(packed_obj)
        return packed_obj

    @classmethod
    def from_json(cls, data: Dict) -> Rectangle:
        """
        Convert a json dict to Rectangle. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: Rectangle in json format as a dict.
        :type data: dict
        :returns: Rectangle from json.
        :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                figure_json = {
                    "points": {
                        "exterior": [
                            [100, 100],
                            [900, 700]
                        ],
                        "interior": []
                    }
                }
                figure = sly.Rectangle.from_json(figure_json)
        """
        validation.validate_geometry_points_fields(data)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)

        exterior = data[POINTS][EXTERIOR]
        if len(exterior) != 2:
            raise ValueError(
                '"exterior" field must contain exactly two points to create Rectangle object.'
            )
        [top, bottom] = sorted([exterior[0][1], exterior[1][1]])
        [left, right] = sorted([exterior[0][0], exterior[1][0]])

        return cls(
            top=top,
            left=left,
            bottom=bottom,
            right=right,
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    def crop(self, other: Rectangle) -> List[Rectangle]:
        """
        Crops current Rectangle.

        :param other: Rectangle to crop from.
        :type other: :class:`~supervisely.geometry.rectangle.Rectangle`
        :returns: List of Rectangles from Rectangle.
        :rtype: List[:class:`~supervisely.geometry.rectangle.Rectangle`]

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                crop_figures = figure.crop(sly.Rectangle(0, 0, 300, 350))
        """
        top = max(self.top, other.top)
        left = max(self.left, other.left)
        bottom = min(self.bottom, other.bottom)
        right = min(self.right, other.right)
        is_valid = (bottom >= top) and (left <= right)
        return [Rectangle(top=top, left=left, bottom=bottom, right=right)] if is_valid else []

    def _transform(self, transform_fn):
        """ """
        transformed_corners = [transform_fn(p) for p in self.corners]
        rows, cols = zip(*points_to_row_col_list(transformed_corners))
        return Rectangle(
            top=round(min(rows)),
            left=round(min(cols)),
            bottom=round(max(rows)),
            right=round(max(cols)),
        )

    @property
    def corners(
        self,
    ) -> List[PointLocation, PointLocation, PointLocation, PointLocation]:
        """
        Get list of Rectangle corners.

        :returns: List of Rectangle corners.
        :rtype: List[:class:`~supervisely.geometry.point_location.PointLocation`]

        :Usage Example:

            .. code-block:: python

                corners = figure.corners
                for corner in corners:
                    print(corner.row, corner.col)
                # Output:
                # 100 100
                # 100 900
                # 700 900
                # 700 100
        """
        return [
            PointLocation(row=self.top, col=self.left),
            PointLocation(row=self.top, col=self.right),
            PointLocation(row=self.bottom, col=self.right),
            PointLocation(row=self.bottom, col=self.left),
        ]

    def rotate(self, rotator: ImageRotator) -> Rectangle:
        """
        Rotates current Rectangle.

        :param rotator: Class for object rotation.
        :type rotator: :class:`~supervisely.geometry.image_rotator.ImageRotator`
        :returns: Rotated Rectangle.
        :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`

        :Usage Example:

            .. code-block:: python

                from supervisely.geometry.image_rotator import ImageRotator

                # Remember that Rectangle class object is immutable, and we need to assign new instance of Rectangle to a new variable
                height, width = 300, 400
                rotator = ImageRotator((height, width), 25)
                rotate_figure = figure.rotate(rotator)
        """
        return self._transform(lambda p: rotator.transform_point(p))

    def resize(self, in_size: Tuple[int, int], out_size: Tuple[int, int]) -> Rectangle:
        """
        Resizes current Rectangle.

        :param in_size: Input image size (height, width) to which belongs Rectangle.
        :type in_size: Tuple[int, int]
        :param out_size: Desired output image size (height, width) to which belongs Rectangle.
        :type out_size: Tuple[int, int]
        :returns: Resized Rectangle.
        :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`

        :Usage Example:

            .. code-block:: python

                # Remember that Rectangle class object is immutable, and we need to assign new instance of Rectangle to a new variable
                in_height, in_width = 300, 400
                out_height, out_width = 600, 800
                resize_figure = figure.resize((in_height, in_width), (out_height, out_width))
        """
        return self._transform(lambda p: p.resize(in_size, out_size))

    def scale(self, factor: float) -> Rectangle:
        """
        Scales current Rectangle.

        :param factor: Scale parameter.
        :type factor: float
        :returns: Scaled Rectangle.
        :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`

        :Usage Example:

            .. code-block:: python

                # Remember that Rectangle class object is immutable, and we need to assign new instance of Rectangle to a new variable
                scale_figure = figure.scale(0.75)
        """
        return self._transform(lambda p: p.scale(factor))

    def translate(self, drow: int, dcol: int) -> Rectangle:
        """
        Translates current Rectangle.

        :param drow: Horizontal shift.
        :type drow: int
        :param dcol: Vertical shift.
        :type dcol: int
        :returns: Translated Rectangle.
        :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`

        :Usage Example:

            .. code-block:: python

                # Remember that Rectangle class object is immutable, and we need to assign new instance of Rectangle to a new variable
                translate_figure = figure.translate(150, 250)
        """
        return self._transform(lambda p: p.translate(drow, dcol))

    def fliplr(self, img_size: Tuple[int, int]) -> Rectangle:
        """
        Flips current Rectangle in horizontal.

        :param img_size: Input image size (height, width) to which belongs Rectangle.
        :type img_size: Tuple[int, int]
        :returns: Flipped Rectangle in horizontal.
        :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`

        :Usage Example:

            .. code-block:: python

                # Remember that Rectangle class object is immutable, and we need to assign new instance of Rectangle to a new variable
                height, width = 300, 400
                fliplr_figure = figure.fliplr((height, width))
        """
        img_width = img_size[1]
        return Rectangle(
            top=self.top,
            left=(img_width - self.right),
            bottom=self.bottom,
            right=(img_width - self.left),
        )

    def flipud(self, img_size: Tuple[int, int]) -> Rectangle:
        """
        Flips current Rectangle in vertical.

        :param img_size: Input image size (height, width) to which belongs Rectangle.
        :type img_size: Tuple[int, int]
        :returns: Flipped Rectangle in vertical.
        :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`

        :Usage Example:

            .. code-block:: python

                # Remember that Rectangle class object is immutable, and we need to assign new instance of Rectangle to a new variable
                height, width = 300, 400
                flipud_figure = figure.flipud((height, width))
        """
        img_height = img_size[0]
        return Rectangle(
            top=(img_height - self.bottom),
            left=self.left,
            bottom=(img_height - self.top),
            right=self.right,
        )

    def _draw_impl(self, bitmap: np.ndarray, color, thickness=1, config=None):
        """ """
        self._draw_contour_impl(bitmap, color, thickness=cv2.FILLED, config=config)  # due to cv2

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        """ """
        cv2.rectangle(
            bitmap,
            pt1=(self.left, self.top),
            pt2=(self.right, self.bottom),
            color=color,
            thickness=thickness,
        )

    def to_bbox(self) -> Rectangle:
        """
        Makes a copy of Rectangle.

        :returns: Copy of Rectangle.
        :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`

        :Usage Example:

            .. code-block:: python

                # Remember that Rectangle class object is immutable, and we need to assign new instance of Rectangle to a new variable
                new_figure = figure.to_bbox()
        """
        return self.clone()

    @property
    def area(self) -> float:
        """
        Rectangle area.

        :returns: Area of current Rectangle.
        :rtype: float

        :Usage Example:

            .. code-block:: python

                print(figure.area)
                # Output: 7288.0
        """
        return float(self.width * self.height)

    @classmethod
    def from_array(cls, arr: np.ndarray) -> Rectangle:
        """
        Create Rectangle with given array shape.

        :param arr: Numpy array.
        :type arr: np.ndarray
        :returns: Rectangle from array.
        :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                np_array = np.zeros((300, 400))
                figure_from_np = sly.Rectangle.from_array(np_array)
        """
        return cls(top=0, left=0, bottom=arr.shape[0] - 1, right=arr.shape[1] - 1)

    # TODO re-evaluate whether we need this, looks trivial.
    @classmethod
    def from_size(cls, size: Tuple[int, int]) -> Rectangle:
        """
        Create Rectangle with given size shape.

        :param size: Input size.
        :type size: Tuple[int, int]
        :returns: Rectangle from size.
        :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                size = (300, 400)
                figure_from_size = sly.Rectangle.from_size(size)
        """
        return cls(0, 0, size[0] - 1, size[1] - 1)

    @classmethod
    def from_geometries_list(cls, geometries: List[Geometry]) -> Rectangle:
        """
        Create Rectangle from given geometry objects.

        :param geometries: List of geometry type objects: :class:`~supervisely.geometry.bitmap.Bitmap>`, :class:`~supervisely.geometry.cuboid.Cuboid>`, :class:`~supervisely.geometry.point.Point>`, :class:`~supervisely.geometry.polygon.Polygon>`, :class:`~supervisely.geometry.polyline.Polyline>`, :class:`~supervisely.geometry.rectangle.Rectangle>`, :class:`~supervisely.geometry.graph.GraphNodes>`.
        :type geometries: List[:class:`~supervisely.geometry.geometry.Geometry`]
        :returns: Rectangle from given geometry objects.
        :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                geom_objs = [sly.Point(100, 200), sly.Polyline([sly.PointLocation(730, 2104), sly.PointLocation(2479, 402)])]
                figure_from_geom_objs = sly.Rectangle.from_geometries_list(geom_objs)
        """
        if geometries is None or len(geometries) == 0:
            raise ValueError("No geometries provided to create a Rectangle.")
        bboxes = [g.to_bbox() for g in geometries]
        top = min(bbox.top for bbox in bboxes)
        left = min(bbox.left for bbox in bboxes)
        bottom = max(bbox.bottom for bbox in bboxes)
        right = max(bbox.right for bbox in bboxes)
        return cls(top=top, left=left, bottom=bottom, right=right)

    @property
    def left(self) -> int:
        """
        Minimal horizontal value of Rectangle.

        :returns: Minimal horizontal value
        :rtype: int

        :Usage Example:

            .. code-block:: python

                print(figure.left)
                # Output: 100
        """
        return self._points[0].col

    @property
    def right(self) -> int:
        """
        Maximal horizontal value of Rectangle.

        :returns: Maximal horizontal value
        :rtype: int

        :Usage Example:

            .. code-block:: python

                print(figure.right)
                # Output: 900
        """
        return self._points[1].col

    @property
    def top(self) -> int:
        """
        Minimal vertical value of Rectangle.

        :returns: Minimal vertical value
        :rtype: int

        :Usage Example:

            .. code-block:: python

                print(rectangle.top)
                # Output: 100
        """
        return self._points[0].row

    @property
    def bottom(self) -> int:
        """
        Maximal vertical value of Rectangle.

        :returns: Maximal vertical value
        :rtype: int

        :Usage Example:

            .. code-block:: python

                print(figure.bottom)
                # Output: 700
        """
        return self._points[1].row

    @property
    def center(self) -> PointLocation:
        """
        Center of Rectangle.

        :returns: Center of Rectangle.
        :rtype: :class:`~supervisely.geometry.point_location.PointLocation`

        :Usage Example:

            .. code-block:: python

                center = figure.center()
        """
        return PointLocation(row=(self.top + self.bottom) // 2, col=(self.left + self.right) // 2)

    @property
    def width(self) -> int:
        """
        Width of Rectangle.

        :returns: Width
        :rtype: int

        :Usage Example:

            .. code-block:: python

                print(figure.width)
                # Output: 801
        """
        return self.right - self.left + 1

    @property
    def height(self) -> int:
        """
        Height of Rectangle

        :returns: Height
        :rtype: int

        :Usage Example:

            .. code-block:: python

                print(figure.height)
                # Output: 601
        """
        return self.bottom - self.top + 1

    def contains(self, rect: Rectangle) -> bool:
        """
        Checks if Rectangle contains a given Rectangle object.

        :param rect: Rectangle to check if it contains.
        :type rect: :class:`~supervisely.geometry.rectangle.Rectangle`
        :returns: True if Rectangle contains given Rectangle, otherwise False
        :rtype: bool

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                rect = sly.Rectangle(200, 250, 400, 500))
                print(figure.contains(rect))
                # Output: True
        """
        return (
            self.left <= rect.left
            and self.right >= rect.right
            and self.top <= rect.top
            and self.bottom >= rect.bottom
        )

    def contains_point_location(self, pt: PointLocation) -> bool:
        """
        Checks if Rectangle contains a given PointLocation object.

        :param pt: PointLocation to check if it contains.
        :type pt: :class:`~supervisely.geometry.point_location.PointLocation`
        :returns: True if Rectangle contains given PointLocation, otherwise False
        :rtype: bool

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                pt = sly.PointLocation(250, 300))
                print(figure.contains_point_location(pt))
                # Output: True
        """
        return (self.left <= pt.col <= self.right) and (self.top <= pt.row <= self.bottom)

    def to_size(self) -> Tuple[int, int]:
        """
        Height and width of Rectangle.

        :returns: Height and width of Rectangle.
        :rtype: :class:`Tuple[int, int]`

        :Usage Example:

            .. code-block:: python

                height, width = figure.to_size()
                print(height, width)
                # Output: 700 900
        """
        return self.height, self.width

    def get_cropped_numpy_slice(self, data: np.ndarray) -> np.ndarray:
        """
        Slice of given numpy array with Rectangle.

        :param data: Numpy array.
        :type data: np.ndarray
        :returns: Sliced numpy array
        :rtype: :class:`np.ndarray<np.ndarray>`

        :Usage Example:

            .. code-block:: python

                np_slice = np.zeros((200, 500))
                mask_slice = figure.get_cropped_numpy_slice(np_slice)
                print(mask_slice.shape)
                # Output: (199, 499)
        """
        return data[self.top : (self.bottom + 1), self.left : (self.right + 1), ...]

    def intersects_with(self, rect: Rectangle) -> bool:
        """
        Checks intersects Rectangle with given Rectangle object or not.

        :param rect: Rectangle to check if it intersects with.
        :type rect: :class:`~supervisely.geometry.rectangle.Rectangle`
        :returns: True if given Rectangle intersects with Rectangle, otherwise False
        :rtype: bool

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                rect = sly.Rectangle(90, 90, 400, 500)
                print(figure.intersects_with(rect))
                # Output: True
        """
        if self.left > rect.right or self.right < rect.left:
            return False
        if self.top > rect.bottom or self.bottom < rect.top:
            return False
        return True

    @classmethod
    def allowed_transforms(cls):
        """
        Returns the allowed transforms for the Rectangle.
        """
        from supervisely.geometry.alpha_mask import AlphaMask
        from supervisely.geometry.any_geometry import AnyGeometry
        from supervisely.geometry.bitmap import Bitmap
        from supervisely.geometry.polygon import Polygon
        from supervisely.geometry.oriented_bbox import OrientedBBox

        return [AlphaMask, AnyGeometry, Bitmap, Polygon, Rectangle, OrientedBBox]

    @classmethod
    def _round_subpixel_coordinates(
        cls,
        top: Union[int, float],
        left: Union[int, float],
        bottom: Union[int, float],
        right: Union[int, float],
    ) -> Tuple[int, int, int, int]:
        """
        Apply rounding logic to the Rectangle coordinates. Coordinates will remain in subpixel precision.

        Top will be rounded down if the decimal part is lesser than 0.7, it will include the vertical pixel if it is in the range, otherwise it will be rounded up and this pixel will not be included.
        Left will be rounded down if the decimal part is lesser than 0.7, it will include the horizontal pixel if it is in the range, otherwise it will be rounded up and this pixel will not be included.
        Bottom will be rounded up if the decimal part is greater than 0.3, it will include the vertical pixel if it is in the range, otherwise it will be rounded down and this pixel will not be included.
        Right will be rounded up if the decimal part is greater than 0.3, it will include the horizontal pixel if it is in the range, otherwise it will be rounded down and this pixel will not be included.

        Example:
        Input coordinates:
        - top = 1.55, left = 1.74, bottom = 4.63, right = 3.76
        Output coordinates:
        - top = 1, left = 2, bottom = 5, right = 4
        - top will be rounded down to 2, left will be rounded down to 2, bottom will be rounded up to 6, right will be rounded down to 6


        Subpixel coordinate system:
            0   1   2   3   4   5
        0   +---+---+---+---+---+
            |   |   |   |   |   |
        1   +---+---+---+---+---+
            |   |   | x | x |   |
        2   +---+---+---+---+---+
            |   |   | x | x |   |
        3   +---+---+---+---+---+
            |   |   | x | x |   |
        4   +---+---+---+---+---+
            |   |   | x | x |   |
        5   +---+---+---+---+---+
                      x   x

        :param top: Minimal vertical value of Rectangle.
        :type top: Union[int, float]
        :param left: Minimal horizontal value of Rectangle.
        :type left: Union[int, float]
        :param bottom: Maximal vertical value of Rectangle.
        :type bottom: Union[int, float]
        :param right: Maximal horizontal value of Rectangle.
        :type right: Union[int, float]
        :returns: Rounded Rectangle coordinates
        :rtype: Tuple[int, int, int, int]
        """
        RIGHT_OVERLAP = 0.3
        LEFT_OVERLAP = 1 - RIGHT_OVERLAP

        # Check case if all coords in 1 pixel range
        if int(top) == int(bottom) and int(left) == int(right):
            return int(top), int(left), int(bottom), int(right)

        if top % 1 >= LEFT_OVERLAP:
            top = ceil(top)
        else:
            top = floor(top)

        if left % 1 >= LEFT_OVERLAP:
            left = ceil(left)
        else:
            left = floor(left)

        if bottom % 1 >= RIGHT_OVERLAP:
            bottom = ceil(bottom)
        else:
            bottom = floor(bottom)
        if right % 1 >= RIGHT_OVERLAP:
            right = ceil(right)
        else:
            right = floor(right)
        return top, left, bottom, right

    @classmethod
    def _to_pixel_coordinate_system_json(cls, data: Dict, image_size: List[int]) -> Dict:
        """
        Convert Rectangle from subpixel precision to pixel precision by subtracting a subpixel offset from the coordinates.

        Points order in json format: [[left, top], [right, bottom]]

        In the labeling tool, labels are created with subpixel precision,
        which means that the coordinates of the rectangle corners (top, left and bottom, right) can have decimal values representing fractions of a pixel.
        However, in Supervisely SDK, geometry coordinates are represented using pixel precision, where the coordinates are integers representing whole pixels.

        Example:
        Step 1. Input coordinates:
        - top = 1.55, left = 1.74, bottom = 4.63, right = 3.76

        Step 2. Round the coordinates (still remain in subpixel precision):
        - top = 1, left = 2, bottom = 5, right = 4
        - top will be rounded down to 2, left will be rounded down to 2, bottom will be rounded up to 6, right will be rounded down to 6

        Draw coordinates in pixel coordinate system:
            0   1   2   3   4   5
        0   +---+---+---+---+---+
            |   |   |   |   |   |
        1   +---+---+---+---+---+
            |   |   | x | x |   |
        2   +---+---+---+---+---+
            |   |   | x | x |   |
        3   +---+---+---+---+---+
            |   |   | x | x |   |
        4   +---+---+---+---+---+
            |   |   | x | x |   |
        5   +---+---+---+---+---+
                      x   x

        Step 3. Convert to pixel coordinates by subtracting a subpixel offset:
        - top = 1, left = 2, bottom = 4, right = 3

        Draw coordinates in pixel coordinate system:
            0   1   2   3   4   5
        0   +---+---+---+---+---+
            |   |   |   |   |   |
        1   +---+---+---+---+---+
            |   |   | x | x |   |
        2   +---+---+---+---+---+
            |   |   | x | x |   |
        3   +---+---+---+---+---+
            |   |   | x | x |   |
        4   +---+---+---+---+---+
            |   |   | x | x |   |
        5   +---+---+---+---+---+

        :param data: Json data with geometry config.
        :type data: dict
        :param image_size: Image size in pixels (height, width).
        :type image_size: List[int]
        :returns: Json data with coordinates converted to pixel coordinate system.
        :rtype: dict
        """
        data = deepcopy(data)  # Avoid modifying the original data
        height, width = image_size[:2]

        exterior = data[POINTS][EXTERIOR]
        [top, bottom] = sorted([exterior[0][1], exterior[1][1]])
        [left, right] = sorted([exterior[0][0], exterior[1][0]])

        top, left, bottom, right = cls._round_subpixel_coordinates(top, left, bottom, right)
        right = max(left, right - 1)
        bottom = max(top, bottom - 1)
        data[POINTS][EXTERIOR] = [[left, top], [right, bottom]]
        return data

    @classmethod
    def _to_subpixel_coordinate_system_json(cls, data: Dict) -> Dict:
        """
        Convert Rectangle from pixel precision to subpixel precision by adding a subpixel offset to the coordinates.

        Points order in json format: [[left, top], [right, bottom]]

        In the labeling tool, labels are created with subpixel precision,
        which means that the coordinates of the rectangle corners (top, left and bottom, right) can have decimal values representing fractions of a pixel.
        However, in Supervisely SDK, geometry coordinates are represented using pixel precision, where the coordinates are integers representing whole pixels.

        :param data: Json data with geometry config.
        :type data: dict
        :returns: Json data with coordinates converted to subpixel coordinate system.
        :rtype: dict
        """
        data = deepcopy(data)  # Avoid modifying the original data

        exterior = data[POINTS][EXTERIOR]
        [top, bottom] = sorted([exterior[0][1], exterior[1][1]])
        [left, right] = sorted([exterior[0][0], exterior[1][0]])

        right = max(left, right + 1)
        bottom = max(top, bottom + 1)
        data[POINTS][EXTERIOR] = [[left, top], [right, bottom]]
        return data

    # def _to_pixel_coordinate_system(self) -> Rectangle:
    #     """
    #     Convert Rectangle from subpixel precision to pixel precision by subtracting a subpixel offset from the coordinates.

    #     In the labeling tool, labels are created with subpixel precision,
    #     which means that the coordinates of the rectangle corners (top, left and bottom, right) can have decimal values representing fractions of a pixel.
    #     However, in Supervisely SDK, geometry coordinates are represented using pixel precision, where the coordinates are integers representing whole pixels.

    #     Example:
    #     Step 1. Input coordinates:
    #     - top = 1.55, left = 1.74, bottom = 4.63, right = 3.76

    #     Step 2. Round the coordinates:
    #     - top = 1, left = 2, bottom = 5, right = 4
    #     - top will be rounded down to 2, left will be rounded down to 2, bottom will be rounded up to 6, right will be rounded down to 6

    #     Draw coordinates in pixel coordinate system:
    #         0   1   2   3   4   5
    #     0   +---+---+---+---+---+
    #         |   |   |   |   |   |
    #     1   +---+---+---+---+---+
    #         |   |   | x | x |   |
    #     2   +---+---+---+---+---+
    #         |   |   | x | x |   |
    #     3   +---+---+---+---+---+
    #         |   |   | x | x |   |
    #     4   +---+---+---+---+---+
    #         |   |   | x | x |   |
    #     5   +---+---+---+---+---+
    #                   x   x

    #     Step 3. Convert to pixel coordinates by subtracting a subpixel offset:
    #     - top = 1, left = 2, bottom = 4, right = 3

    #     Draw coordinates in pixel coordinate system:
    #         0   1   2   3   4   5
    #     0   +---+---+---+---+---+
    #         |   |   |   |   |   |
    #     1   +---+---+---+---+---+
    #         |   |   | x | x |   |
    #     2   +---+---+---+---+---+
    #         |   |   | x | x |   |
    #     3   +---+---+---+---+---+
    #         |   |   | x | x |   |
    #     4   +---+---+---+---+---+
    #         |   |   | x | x |   |
    #     5   +---+---+---+---+---+

    #     :returns: New instance of Rectangle object with corners in pixel format.
    #     :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`
    #     """
    #     left = floor(self.left)
    #     top = floor(self.top)
    #     right = max(left, floor(self.right) - 1)
    #     bottom = max(top, floor(self.bottom) - 1)
    #     return Rectangle(
    #         top=top,
    #         left=left,
    #         bottom=bottom,
    #         right=right,
    #         sly_id=self.sly_id,
    #         class_id=self.class_id,
    #         labeler_login=self.labeler_login,
    #         updated_at=self.updated_at,
    #         created_at=self.created_at,
    #     )

    # def _to_subpixel_coordinate_system(self) -> Rectangle:
    #     """
    #     Convert Rectangle from pixel precision to subpixel precision by adding a subpixel offset to the coordinates.

    #     In the labeling tool, labels are created with subpixel precision,
    #     which means that the coordinates of the rectangle corners (top, left and bottom, right) can have decimal values representing fractions of a pixel.
    #     However, in Supervisely SDK, geometry coordinates are represented using pixel precision, where the coordinates are integers representing whole pixels.

    #     :returns: New instance of Rectangle object with corners in subpixel format.
    #     :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`
    #     """
    #     left = self.left
    #     top = self.top
    #     right = self.right + 1
    #     bottom = self.bottom + 1

    #     return Rectangle(
    #         top=top,
    #         left=left,
    #         bottom=bottom,
    #         right=right,
    #         sly_id=self.sly_id,
    #         class_id=self.class_id,
    #         labeler_login=self.labeler_login,
    #         updated_at=self.updated_at,
    #         created_at=self.created_at,
    #     )



from typing import Dict, Optional

from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget


class RandomSplitsTable(Widget):
    """Table showing train/val/total counts and percentages for a random split; supports adjusting train percent."""

    def __init__(
        self,
        items_count: int,
        start_train_percent: Optional[int] = 80,
        disabled: Optional[bool] = False,
        widget_id: Optional[int] = None,
    ):
        """:param items_count: Total number of items.
        :type items_count: int
        :param start_train_percent: Initial train split percent (1-99).
        :type start_train_percent: int, optional
        :param disabled: If True, widget is disabled.
        :type disabled: bool, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: int, optional

        :raises ValueError: If start_train_percent not in [1, 99].
        """
        self._disabled = disabled
        if 1 <= start_train_percent <= 99:
            pass
        else:
            raise ValueError("start_train_percent must be in range [1; 99].")
        self._table_data = [
            {"name": "train", "type": "success"},
            {"name": "val", "type": "primary"},
            {"name": "total", "type": "gray"},
        ]
        if items_count is None:
            items_count = 0
        self._items_count = items_count
        train_count = int(items_count / 100 * start_train_percent)
        self._count = {"total": items_count, "train": train_count, "val": items_count - train_count}

        self._percent = {
            "total": 100,
            "train": start_train_percent,
            "val": 100 - start_train_percent,
        }

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "table_data": self._table_data,
            "items_count": self._items_count,
            "disabled": self._disabled,
        }

    def get_json_state(self):
        return {"count": self._count, "percent": self._percent}

    def get_splits_counts(self) -> Dict[str, int]:
        return StateJson()[self.widget_id]["count"]

    def set_items_count(self, items_count: int):
        self._items_count = items_count
        current_train_percent = self._percent["train"]
        train_count = int(items_count * current_train_percent / 100)
        self._count = {"total": items_count, "train": train_count, "val": items_count - train_count}

        DataJson()[self.widget_id]["items_count"] = self._items_count
        DataJson().send_changes()
        StateJson()[self.widget_id]["count"] = self._count
        StateJson().send_changes()

    def set_train_split_percent(self, percent: int):
        if 1 <= percent <= 99:
            self._percent["train"] = percent
            self._percent["val"] = 100 - percent
            self._count["train"] = int(self._items_count / 100 * percent)
            self._count["val"] = self._items_count - self._count["train"]

            StateJson()[self.widget_id]["count"] = self._count
            StateJson()[self.widget_id]["percent"] = self._percent
            StateJson().send_changes()
        else:
            raise ValueError("percent must be in range [1; 99].")

    def get_train_split_percent(self) -> int:
        return StateJson()[self.widget_id]["percent"]["train"]

    def set_val_split_percent(self, percent: int):
        if 1 <= percent <= 99:
            self._percent["val"] = percent
            self._percent["train"] = 100 - percent
            self._count["val"] = int(self._items_count / 100 * percent)
            self._count["train"] = self._items_count - self._count["val"]
            StateJson()[self.widget_id]["count"] = self._count
            StateJson()[self.widget_id]["percent"] = self._percent
            StateJson().send_changes()
        else:
            raise ValueError("percent must be in range [1; 99].")

    def get_val_split_percent(self) -> int:
        return StateJson()[self.widget_id]["percent"]["val"]



# coding: utf-8
from supervisely.geometry.constants import EXTERIOR, INTERIOR, POINTS


def _is_2d_typed_coords_valid(coords, individual_coord_types):
    """
    Validate 2D coordinates.

    :param coords: List of 2D coordinates.
    :type coords: list
    :param individual_coord_types: Types of individual coordinates.
    :type individual_coord_types: tuple
    :returns: True if validation successful, False otherwise.
    :rtype: bool
    """
    return isinstance(coords, (list, tuple)) and all(
        len(point) == 2 and all(isinstance(coord, individual_coord_types) for coord in point)
        for point in coords
    )


def _is_2d_numeric_coords_valid(coords) -> bool:
    """
    :param coords: list or tuple of 2 numbers.
    :type coords: list or tuple
    :returns: True if validation successful, False otherwise.
    :rtype: bool
    """
    return _is_2d_typed_coords_valid(coords, (int, float))


def is_2d_int_coords_valid(coords) -> bool:
    """
    :param coords: list (or tuple) of 2 integers.
    :type coords: list or tuple
    :returns: True if validation successful, False otherwise.
    :rtype: bool
    """
    return _is_2d_typed_coords_valid(coords, (int,))


def validate_geometry_points_fields(json_obj: dict) -> None:
    """
    Validate json geometry points container structure, which presented as python dict.
    :param json_obj: example: { "points": { "exterior": [[1,2], [10, 20]], "interior": [[2,3]] } }
    :type json_obj: dict
    :returns: None
    :rtype: None
    """
    if POINTS not in json_obj:
        raise ValueError("Input data must contain {} field.".format(POINTS))

    points_obj = json_obj[POINTS]
    if not isinstance(points_obj, dict):
        raise TypeError('Input data field "{}" must be dict object.'.format(POINTS))

    if EXTERIOR not in points_obj or INTERIOR not in points_obj:
        raise ValueError(
            '"{}" field must contain {} and {} fields.'.format(POINTS, EXTERIOR, INTERIOR)
        )

    if not _is_2d_numeric_coords_valid(points_obj[EXTERIOR]):
        raise TypeError("{} field must be a list of 2 numbers lists.".format(EXTERIOR))

    interior = points_obj[INTERIOR]
    if not isinstance(interior, list) or not all(
        _is_2d_numeric_coords_valid(interior_component) for interior_component in interior
    ):
        raise TypeError("{} field must be a list of lists of 2 numbers lists.".format(INTERIOR))



# coding: utf-8
"""Work with volume objects via the Supervisely API."""

from supervisely.api.entity_annotation.object_api import ObjectApi
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.volume_annotation.volume_object_collection import (
    VolumeObjectCollection,
)


class VolumeObjectApi(ObjectApi):
    """
    API for working with :class:`~supervisely.volume_annotation.volume_object.VolumeObject` in
    :class:`~supervisely.volume_annotation.volume_annotation.VolumeAnnotation`.
    :class:`~supervisely.api.volume.volume_object_api.VolumeObjectApi` object is immutable.
    """

    def append_bulk(
        self, volume_id: int, objects: VolumeObjectCollection, key_id_map: KeyIdMap = None
    ):
        """
        Add objects to VolumeAnnotation.

        :param volume_id: Volume ID in Supervidely.
        :type volume_id: int
        :param objects: VolumeAnnotation objects.
        :type objects: :class:`~supervisely.volume_annotation.volume_object_collection.VolumeObjectCollection`
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :returns: List of objects IDs
        :rtype: List[int]

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                from supervisely.volume_annotation.volume_tag import VolumeTag
                from supervisely.video_annotation.key_id_map import KeyIdMap

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                project_id = 17209
                volume_id = 19402023

                meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(meta_json)

                key_id_map = KeyIdMap()
                ann_info = api.volume.annotation.download(volume_id)
                ann = sly.VolumeAnnotation.from_json(ann_info, project_meta, key_id_map)

                api.volume.object.append_bulk(volume_id, ann.objects, key_id_map)
        """

        info = self._api.volume.get_info_by_id(volume_id)
        return self._append_bulk(
            self._api.volume.tag,
            volume_id,
            info.project_id,
            info.dataset_id,
            objects,
            key_id_map,
        )



# coding: utf-8

import json
import os
import urllib.parse
import uuid
from typing import Dict, List, NamedTuple, Optional

from supervisely.api.module_api import ApiField, ModuleApiBase
from supervisely.collection.str_enum import StrEnum


class NotificationType(StrEnum):
    """Notification severity/type values used in report widgets."""

    INFO = "info"
    """"""
    NOTE = "note"
    """"""
    WARNING = "warning"
    """"""
    ERROR = "error"
    """"""


# @TODO: standardise title/description/name and so on for everyone the same way
class ReportApi(ModuleApiBase):
    """API for working with Reports."""

    def __init__(self, api):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`

        :Usage Example:

            .. code-block:: python

                report = api.report
        """
        ModuleApiBase.__init__(self, api)

    # https://developer.mozilla.org/en-US/docs/Web/CSS/grid-template
    # grid-template: "a a a" 40px "b c c" 40px "b c c" 40px / 1fr 1fr 1fr;
    # area -a or b or c
    def create(self, team_id: int, name: str, widgets, layout: Optional[str] = ""):
        """
        Creates report in the given Team.

        :param team_id: Team ID in Supervisely, where report will be created
        :type team_id: int
        :param name: Report name.
        :type name: str
        :param widgets:
        :type widgets:
        :param layout:
        :type layout:
        :returns:
        :rtype:
        """
        data = {
            ApiField.TEAM_ID: team_id,
            ApiField.NAME: name,
            ApiField.WIDGETS: widgets,
            ApiField.LAYOUT: layout,
        }
        response = self._api.post("reports.create", data)
        return response.json()[ApiField.ID]

    # def create_table(self, df, name, subtitle, per_page=20, pageSizes=[10, 20, 50, 100, 500], fix_columns=None):
    #     res = {
    #         "name": name,
    #         "subtitle": subtitle,
    #         "type": str(WidgetType.TABLE),
    #         "content": json.loads(df.to_json(orient='split')),
    #         "options": {
    #             "perPage": per_page,
    #             "pageSizes": pageSizes,
    #         }
    #     }
    #     if fix_columns is not None:
    #         res["options"]["fixColumns"] = fix_columns
    #     return res
    #
    # def create_notification(self, name, content, notification_type=NotificationType.INFO):
    #     return {
    #         "type": str(WidgetType.NOTIFICATION),
    #         "title": name,
    #         "content": content,
    #         "options": {
    #             "type": str(notification_type)
    #         }
    #     }
    #
    # def create_plotly(self, data_json, name, subtitle):
    #     data = data_json
    #     if type(data) is str:
    #         data = json.loads(data_json)
    #     elif type(data) is not dict:
    #         raise RuntimeError("type(data_json) is not dict")
    #     return {
    #         "name": name,
    #         "subtitle": subtitle,
    #         "type": str(WidgetType.PLOTLY),
    #         "content": data
    #     }
    #
    #
    # def create_linechart(self, name, description, id=None):
    #     res = {
    #         "type": str(WidgetType.LINECHART),
    #         "name": "linechart block title",
    #         "subtitle": "linechart block description",
    #         "content": [],
    #         "options": {}
    #     }
    #     res["id"] = uuid.uuid4().hex if id is None else id
    #     return res

    def url(self, id: int) -> str:
        """
        Get Report URL by ID.

        :param id: Report ID.
        :type id: int
        :returns: Report URL
        :rtype: str
        """
        return urllib.parse.urljoin(self._api.server_address, "reports/{}".format(id))

    def get_widget(self, report_id: int, widget_id: int):
        """
        Get Widget by ID.

        :param report_id: Report ID.
        :type report_id: int
        :param widget_id: Widget ID.
        :type widget_id: int
        :returns: Widget by report ID and widget ID.
        :rtype: :class:`~supervisely.app.widgets.widget.Widget`
        """
        response = self._api.post(
            "reports.widgets.get", {"reportId": report_id, "widgetId": widget_id}
        )
        return response.json()

    def _change_widget(
        self,
        method,
        report_id,
        widget_id,
        widget_type=None,
        name=None,
        description=None,
        area=None,
        content=None,
        options=None,
    ):
        """ """
        data = dict()
        data[ApiField.ID] = widget_id
        if name is not None:
            data[ApiField.NAME] = name
        if widget_type is not None:
            data[ApiField.TYPE] = widget_type
        if description is not None:
            data[ApiField.SUBTITLE] = description
        if area is not None:
            data[ApiField.AREA] = area
        if content is not None:
            data[ApiField.CONTENT] = content
        if options is not None:
            data[ApiField.OPTIONS] = options
        response = self._api.post(method, {ApiField.REPORT_ID: report_id, ApiField.WIDGET: data})
        return response.json()

    def update_widget(
        self,
        report_id: int,
        widget_id: int,
        name: Optional[str] = None,
        description: Optional[str] = None,
        area=None,
        content=None,
        options=None,
    ):
        """
        Method description

        :param report_id: Report ID.
        :type report_id: int
        :param widget_id: Widget ID.
        :type widget_id: int
        :param name:
        :type name: str
        :param description:
        :type description: str
        :param area:
        :type area:
        :param content:
        :type content:
        :param options:
        :type options:

        :returns: Response JSON.
        :rtype: dict
        """
        return self._change_widget(
            "reports.widgets.update",
            report_id,
            widget_id,
            name,
            description,
            area,
            content,
            options,
        )

    def rewrite_widget(
        self,
        report_id: int,
        widget_id: int,
        widget_type,
        name: str = None,
        description: str = None,
        area=None,
        content=None,
        options=None,
    ):
        """
        Method description

        :param report_id: Report ID.
        :type report_id: int
        :param widget_id: Widget ID.
        :type widget_id: int
        :param widget_type: Widget type.
        :type widget_type:
        :param name:
        :type name: str
        :param description:
        :type description: str
        :param area:
        :type area:
        :param content:
        :type content:
        :param options:
        :type options:

        :returns: Response JSON.
        :rtype: dict
        """
        return self._change_widget(
            "reports.widgets.rewrite",
            report_id,
            widget_id,
            widget_type,
            name,
            description,
            area,
            content,
            options,
        )



from typing import Any, Callable, Dict, List, Optional

import supervisely as sly
from supervisely.app import DataJson
from supervisely.app.content import StateJson
from supervisely.app.widgets import Widget
from supervisely.sly_logger import logger
from supervisely.annotation.tag_meta import TagValueType


class TagsTable(Widget):
    """Table displaying tags from a project.
    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/tables/TagsTable>`_.
    """

    class Routes:
        """Route name constants for this widget."""
        TAG_SELECTED = "tag_selected_cb"

    def __init__(
        self,
        project_meta: Optional[sly.ProjectMeta] = None,
        project_id: Optional[int] = None,
        project_fs: Optional[sly.Project] = None,
        allowed_types: Optional[List[TagValueType]] = None,
        selectable: Optional[bool] = True,
        disabled: Optional[bool] = False,
        widget_id: Optional[str] = None,
    ):
        """Initialize the TagsTable widget.

        :param project_meta: Project meta object from which tags will be taken.
        :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
        :param project_id: Project id from which tags will be taken.
        :type project_id: int
        :param project_fs: Local supervisely project from which tags will be taken.
        :type project_fs: :class:`~supervisely.project.project.Project`
        :param allowed_types: List of allowed value types to be displayed.
        :type allowed_types: List[:class:`~supervisely.annotation.tag_meta.TagValueType`]
        :param selectable: If True, user can select tags from table.
        :type selectable: bool
        :param disabled: If True, the elements in the table will be disabled.
        :type disabled: bool
        :param widget_id: Unique widget identifier.
        :type widget_id: str
        :raises ValueError: If both project_id and project_fs parameters are provided.

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import TagsTable

                tags_table = TagsTable(project_id=123, selectable=True)
        """
        if project_id is not None and project_fs is not None:
            raise ValueError(
                "You can not provide both project_id and project_fs parameters to Tags Table widget."
            )
        self._table_data = []
        self._columns = []
        self._changes_handled = False
        self._global_checkbox = False
        self._checkboxes = []
        self._selectable = selectable
        self._selection_disabled = disabled
        self._loading = False
        self._allowed_types = allowed_types if allowed_types is not None else []
        if project_id is not None:
            self._api = sly.Api()
        else:
            self._api = None
        self._project_id = project_id
        if project_id is not None:
            if project_meta is not None:
                logger.warn(
                    "Both parameters project_id and project_meta were provided to TagsTable widget. Project meta tags taken from remote project and project_meta parameter is ignored."
                )
            project_meta = sly.ProjectMeta.from_json(self._api.project.get_meta(project_id))
        self._project_fs = project_fs
        if project_fs is not None:
            if project_meta is not None:
                logger.warn(
                    "Both parameters project_fs and project_meta were provided to TagsTable widget. Project meta tags taken from project_fs.meta and project_meta parameter is ignored."
                )
            project_meta = project_fs.meta

        self._project_meta = project_meta
        if project_meta is not None:
            self._update_meta(project_meta=project_meta)
        super().__init__(widget_id=widget_id, file_path=__file__)

    def value_changed(self, func: Callable[[List[str]], Any]) -> Callable[[], None]:
        """Decorator for the function to be called when the value of the widget changes.

        :param func: Function to be called when the value of the widget changes.
        :type func: Callable[[List[str]], Any]
        :returns: Decorated function.
        :rtype: Callable[[], None]
        """
        route_path = self.get_route_path(TagsTable.Routes.TAG_SELECTED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _value_changed():
            res = self.get_selected_tags()
            func(res)

        return _value_changed

    def _update_meta(self, project_meta: sly.ProjectMeta) -> None:
        columns = ["tag", "type", "applicable to", "possible values"]
        stats = None
        data_to_show = []
        for tag_meta in project_meta.tag_metas:
            if len(self._allowed_types) == 0 or tag_meta.value_type in self._allowed_types:
                data_to_show.append(tag_meta.to_json())

        if self._project_id is not None:
            stats = self._api.project.get_stats(self._project_id)
            project_info = self._api.project.get_info_by_id(self._project_id)
            if project_info.type == str(sly.ProjectType.IMAGES):
                columns.extend(["images count", "labels count"])
            elif project_info.type == str(sly.ProjectType.VIDEOS):
                columns.extend(["videos count", "figures count"])
            elif project_info.type in [
                str(sly.ProjectType.POINT_CLOUDS),
                str(sly.ProjectType.POINT_CLOUD_EPISODES),
            ]:
                columns.extend(["pointclouds count", "figures count"])
            elif project_info.type == str(sly.ProjectType.VOLUMES):
                columns.extend(["volumes count", "figures count"])

            tag_items = {}
            for item in stats["imageTags"]["items"]:
                tag_items[item["tagMeta"]["name"]] = item["total"]

            tag_objects = {}
            for item in stats["objectTags"]["items"]:
                tag_objects[item["tagMeta"]["name"]] = item["total"]
            for tag_meta in data_to_show:
                tag_meta["itemsCount"] = tag_items[tag_meta["name"]]
                tag_meta["objectsCount"] = tag_objects[tag_meta["name"]]

        elif self._project_fs is not None:
            project_stats = self._project_fs.get_tags_stats()

            if type(self._project_fs) == sly.Project:
                columns.extend(["images count", "labels count"])

            elif type(self._project_fs) == sly.VideoProject:
                columns.extend(["videos count", "objects count", "figures count"])

            elif type(self._project_fs) in [sly.PointcloudProject, sly.PointcloudEpisodeProject]:
                columns.extend(["pointclouds count", "objects count", "figures count"])

            elif type(self._project_fs) == sly.VolumeProject:
                columns.extend(["volumes count", "objects count", "figures count"])

            for tag_meta in data_to_show:
                tag_meta["itemsCount"] = project_stats["items_count"][tag_meta["title"]]
                tag_meta["objectsCount"] = project_stats["objects_count"][tag_meta["title"]]
                if type(self._project_fs) != sly.Project:
                    tag_meta["figuresCount"] = project_stats["figures_count"][tag_meta["title"]]

        columns = [col.upper() for col in columns]
        if data_to_show:
            table_data = []
            if self._project_id is not None or self._project_fs is not None:
                data_to_show = sorted(
                    data_to_show, key=lambda line: line["objectsCount"], reverse=True
                )
            for line in data_to_show:
                table_line = []
                table_line.extend(
                    [
                        {
                            "name": "TAG",
                            "data": line["name"],
                            "color": line["color"],
                        },
                        {"name": "TYPE", "data": line["value_type"]},
                        {"name": "APPLICABLE TO", "data": line["applicable_type"]},
                        {"name": "POSSIBLE VALUES", "data": line.get("values", "-")},
                    ]
                )
                if "itemsCount" in line.keys():
                    table_line.append({"name": "ITEMS COUNT", "data": line["itemsCount"]})
                if "objectsCount" in line.keys():
                    table_line.append({"name": "OBJECTS COUNT", "data": line["objectsCount"]})
                # if "figuresCount" in line.keys():
                # table_line.append({"name": "FIGURES COUNT", "data": line["figuresCount"]})
                table_data.append(table_line)
            self._table_data = table_data
            self._columns = columns
            self._checkboxes = [False] * len(table_data)
            self._global_checkbox = False
        else:
            self._table_data = []
            self._columns = []
            self._checkboxes = []
            self._global_checkbox = False

    def read_meta(self, project_meta: sly.ProjectMeta) -> None:
        """Read project meta and update table data.

        :param project_meta: Project meta object from which tags will be taken.
        :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
        """
        self.loading = True
        self._project_fs = None
        self._project_id = None
        self._project_meta = project_meta
        self.clear_selection()
        self._update_meta(project_meta=project_meta)

        DataJson()[self.widget_id]["table_data"] = self._table_data
        DataJson()[self.widget_id]["columns"] = self._columns
        DataJson().send_changes()
        StateJson()[self.widget_id]["checkboxes"] = self._checkboxes
        StateJson()[self.widget_id]["global_checkbox"] = self._global_checkbox
        StateJson().send_changes()
        self.loading = False

    def read_project(self, project_fs: sly.Project) -> None:
        """Read local project and update table data.

        :param project_fs: :class:`~supervisely.project.project.Project` object from which tags will be taken.
        :type project_fs: :class:`~supervisely.project.project.Project`
        """
        self.loading = True
        self._project_fs = project_fs
        self._project_id = None
        self._project_meta = project_fs.meta
        self.clear_selection()
        self._update_meta(project_meta=project_fs.meta)

        DataJson()[self.widget_id]["table_data"] = self._table_data
        DataJson()[self.widget_id]["columns"] = self._columns
        DataJson().send_changes()
        StateJson()[self.widget_id]["checkboxes"] = self._checkboxes
        StateJson()[self.widget_id]["global_checkbox"] = self._global_checkbox
        StateJson().send_changes()
        self.loading = False

    def read_project_from_id(self, project_id: int) -> None:
        """Read remote project by id and update table data.

        :param project_id: Project id from which tags will be taken.
        :type project_id: int
        """
        self.loading = True
        self._project_fs = None
        self._project_id = project_id
        if self._api is None:
            self._api = sly.Api()
        project_meta = sly.ProjectMeta.from_json(self._api.project.get_meta(project_id))
        self._project_meta = project_meta
        self.clear_selection()
        self._update_meta(project_meta=project_meta)

        DataJson()[self.widget_id]["table_data"] = self._table_data
        DataJson()[self.widget_id]["columns"] = self._columns
        DataJson().send_changes()
        StateJson()[self.widget_id]["checkboxes"] = self._checkboxes
        StateJson()[self.widget_id]["global_checkbox"] = self._global_checkbox
        StateJson().send_changes()
        self.loading = False

    def get_json_data(self) -> Dict[str, Any]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - table_data: List of dictionaries with table data.
            - columns: List of column names.
            - loading: If True, the widget is in loading state.
            - disabled: If True, the elements in the table will be disabled.
            - selectable: If True, user can select tags from table.

        :returns: Dictionary with widget data.
        :rtype: Dict[str, Any]
        """
        return {
            "table_data": self._table_data,
            "columns": self._columns,
            "loading": self._loading,
            "disabled": self._selection_disabled,
            "selectable": self._selectable,
        }

    @property
    def allowed_types(self) -> List[TagValueType]:
        """Returns list of allowed tag value types to be displayed in table.

        :returns: List of allowed tag value types to be displayed in table.
        :rtype: List[:class:`~supervisely.annotation.tag_meta.TagValueType`]
        """
        return self._allowed_types

    @property
    def project_id(self) -> int:
        """Returns project id from which tags was taken.

        :returns: Project id from which tags was taken.
        :rtype: int
        """
        return self._project_id

    @property
    def project_fs(self) -> int:
        """Returns project object from which tags was taken.

        :returns: Local supervisely project from which tags was taken.
        :rtype: :class:`~supervisely.project.project.Project`
        """
        return self._project_fs

    @property
    def loading(self) -> bool:
        """Returns True if the widget is in loading state.

        :returns: True if the widget is in loading state.
        :rtype: bool
        """
        return self._loading

    @property
    def project_meta(self) -> bool:
        """Returns project meta object from which tags was taken.

        :returns: Project meta object from which tags was taken.
        :rtype: :class:`~supervisely.project.project_meta.ProjectMeta`
        """
        return self._project_meta

    @loading.setter
    def loading(self, value: bool) -> None:
        """Sets loading state of the widget.

        :param value: Loading state of the widget.
        :type value: bool
        """
        self._loading = value
        DataJson()[self.widget_id]["loading"] = self._loading
        DataJson().send_changes()

    def get_json_state(self) -> Dict[str, Any]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
            - global_checkbox: State of global checkbox.
            - checkboxes: List of checkboxes states.

        :returns: Dictionary with widget state.
        :rtype: Dict[str, Any]
        """
        return {
            "global_checkbox": self._global_checkbox,
            "checkboxes": self._checkboxes,
        }

    def get_selected_tags(self) -> List[str]:
        """Returns list of selected tags.

        :returns: List of selected tags.
        :rtype: List[str]
        """
        tags = []
        for i, line in enumerate(self._table_data):
            checkboxes = StateJson()[self.widget_id]["checkboxes"]
            if len(checkboxes) == 0:
                checkboxes = [False] * len(self._table_data)
            if checkboxes[i]:
                for col in line:
                    if col["name"] == "TAG":
                        tags.append(col["data"])
        return tags

    def clear_selection(self) -> None:
        """Clears selection of tags."""
        self._global_checkbox = False
        self._checkboxes = [False] * len(self._table_data)
        StateJson()[self.widget_id]["global_checkbox"] = self._global_checkbox
        StateJson()[self.widget_id]["checkboxes"] = self._checkboxes
        StateJson().send_changes()

    def set_project_meta(self, project_meta: sly.ProjectMeta) -> None:
        """Sets project meta object from which tags will be taken.

        :param project_meta: Project meta object from which tags will be taken.
        :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
        """
        self._update_meta(project_meta)
        self._project_meta = project_meta
        self.update_data()
        DataJson().send_changes()

    def select_tags(self, tags: List[str]) -> None:
        """Selects tags in the table from given list.

        :param tags: List of tags to be selected.
        :type tags: List[str]
        """
        self._global_checkbox = False
        self._checkboxes = [False] * len(self._table_data)

        project_tags = []
        for i, line in enumerate(self._table_data):
            for col in line:
                if col["name"] == "TAG":
                    project_tags.append(col["data"])

        for i, cls_name in enumerate(project_tags):
            if cls_name in tags:
                self._checkboxes[i] = True
        self._global_checkbox = all(self._checkboxes)
        StateJson()[self.widget_id]["global_checkbox"] = self._global_checkbox
        StateJson()[self.widget_id]["checkboxes"] = self._checkboxes
        StateJson().send_changes()

    def select_all(self) -> None:
        """Selects all tags in the table."""
        self._global_checkbox = True
        self._checkboxes = [True] * len(self._table_data)
        StateJson()[self.widget_id]["global_checkbox"] = self._global_checkbox
        StateJson()[self.widget_id]["checkboxes"] = self._checkboxes
        StateJson().send_changes()



# coding: utf-8
import os.path

ARTIFACTS = 'artifacts'
CODE = 'code'
CONFIG_JSON = 'config.json'
DATA = 'data'
GRAPH_JSON = 'graph.json'
MODEL = 'model'
PROJECTS = 'projects'
TASK_CONFIG_JSON = 'task_config.json'
TASK_SETTINGS_JSON = 'task_settings.json'
RESULTS = 'results'
TMP = 'tmp'
TASK_SHARED = 'sly_task_shared'


class TaskPaths:
    """Constants for common file and directory paths inside a running Supervisely task environment."""

    TASK_DIR = '/sly_task_data'
    SETTINGS_PATH = os.path.join(TASK_DIR, TASK_SETTINGS_JSON)  # Deprecated - use TASK_CONFIG_PATH instead
    TASK_CONFIG_PATH = os.path.join(TASK_DIR, TASK_CONFIG_JSON)
    DATA_DIR = os.path.join(TASK_DIR, DATA)
    RESULTS_DIR = os.path.join(TASK_DIR, RESULTS)
    DEBUG_DIR = os.path.join(TASK_DIR, TMP)
    GRAPH_PATH = os.path.join(TASK_DIR, GRAPH_JSON)
    MODEL_DIR = os.path.join(TASK_DIR, MODEL)
    MODEL_CONFIG_PATH = os.path.join(MODEL_DIR, CONFIG_JSON)
    MODEL_CONFIG_NAME = CONFIG_JSON

    # For Python tasks only.
    OUT_ARTIFACTS_DIR = os.path.join(RESULTS_DIR, ARTIFACTS)
    OUT_PROJECTS_DIR = os.path.join(RESULTS_DIR, PROJECTS)
    TASK_SHARED_DIR = '/' + TASK_SHARED



class VersionSchemaField:
    """String constants used as column names across snapshot table schemas."""

    SCHEMA_VERSION = "schema_version"
    TABLES = "tables"
    NAME = "name"
    PATH = "path"
    ROW_COUNT = "row_count"
    SRC_DATASET_ID = "src_dataset_id"
    PARENT_SRC_DATASET_ID = "parent_src_dataset_id"
    FULL_PATH = "full_path"
    DESCRIPTION = "description"
    SRC_VIDEO_ID = "src_video_id"
    SRC_VOLUME_ID = "src_volume_id"
    HASH = "hash"
    LINK = "link"
    FRAMES_COUNT = "frames_count"
    FRAME_WIDTH = "frame_width"
    FRAME_HEIGHT = "frame_height"
    FRAMES_TO_TIMECODES = "frames_to_timecodes"
    META = "meta"
    CUSTOM_DATA = "custom_data"
    CREATED_AT = "created_at"
    UPDATED_AT = "updated_at"
    ANN_JSON = "ann_json"
    JSON = "json"
    ANNOTATION = "annotation"
    SRC_OBJECT_ID = "src_object_id"
    CLASS_NAME = "class_name"
    KEY = "key"
    TAGS_JSON = "tags_json"
    SRC_FIGURE_ID = "src_figure_id"
    FRAME_INDEX = "frame_index"
    GEOMETRY_TYPE = "geometry_type"
    GEOMETRY_JSON = "geometry_json"



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from typing import Optional

from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget
from supervisely.app.widgets_context import JinjaWidgets


class RunAppButton(Widget):
    """Button that launches a Supervisely app (module) in team/workspace with optional payload and visibility control."""

    def __init__(
        self,
        team_id: int,
        workspace_id: int,
        module_id: int,
        payload: dict = None,
        text: Optional[str] = "Button",
        button_type: Optional[
            Literal["primary", "info", "warning", "danger", "success", "text"]
        ] = "primary",
        button_size: Optional[Literal["mini", "small", "large"]] = None,
        plain: Optional[bool] = False,
        icon: Optional[str] = None,
        icon_gap: Optional[int] = 5,
        available_in_offline: Optional[bool] = False,
        visible_by_vue_field: Optional[str] = "",
        check_existing_task_cb: Optional[str] = "null",
        widget_id: Optional[str] = None,
    ):
        """
        Button the runs an app on Supervisely instance.

        :param team_id: Team ID.
        :type team_id: int
        :param workspace_id: Workspace ID.
        :type workspace_id: int
        :param module_id: Module ID.
        :type module_id: int
        :param payload: Payload to be sent to the app.
        :type payload: dict, optional
        :param text: Text to be displayed on the button.
        :type text: str, optional
        :param button_type: Type of the button.
        :type button_type: Literal["primary", "info", "warning", "danger", "success", "text"], optional
        :param button_size: Size of the button.
        :type button_size: Literal["mini", "small", "large"], optional
        :param plain: If True, the button will be plain.
        :type plain: bool, optional
        :param icon: Icon to be displayed on the button.
        :type icon: str, optional
        :param icon_gap: Gap between the icon and the text.
        :type icon_gap: int, optional
        :param available_in_offline: If True, the button will be available in offline session.
        :type available_in_offline: bool, optional
        :param visible_by_vue_field: Vue field that controls the button visibility. If set to "isStaticVersion", the button will be visible only in offline session.
        :type visible_by_vue_field: str, optional
        :param check_existing_task_cb: Sets the callback function for checking existing tasks. Function should be a string (docstring) of JavaScript code.
        :type check_existing_task_cb: str, optional
        :param widget_id: Widget ID.
        :type widget_id: str, optional

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import RunAppButton

                workspace_id = 123
                project_id = 555
                app_module_id = 777

                run_app_button = RunAppButton(
                    workspace_id=workspace_id,
                    module_id=app_module_id,
                    payload={"state": {"slyProjectId": project_id}},
                    text="Run App",
                    button_type="info",
                    plain=True,
                    icon="zmdi zmdi-chart",
                    available_in_offline=True,
                    visible_by_vue_field="isStaticVersion",
                )
        """
        self._text = text
        self._button_type = button_type
        self._button_size = button_size
        self._plain = plain
        self._icon_gap = icon_gap
        if icon is None:
            self._icon = ""
        else:
            self._icon = f'<i class="{icon}" style="margin-right: {icon_gap}px"></i>'

        self._available_in_offline = available_in_offline
        self._visible_by_vue_field = visible_by_vue_field
        self._loading = False
        self._disabled = False
        self._team_id = team_id
        self._workspace_id = workspace_id
        self._module_id = module_id
        self._payload = payload
        self._check_existing_task_cb = check_existing_task_cb

        super().__init__(widget_id=widget_id, file_path=__file__)

        script_path = "./sly/css/app/widgets/run_app_button/script.js"
        JinjaWidgets().context["__widget_scripts__"][self.__class__.__name__] = script_path

    def get_json_data(self):
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - text: Text to be displayed on the button.
            - button_type: Type of the button.
            - plain: If True, the button will be plain.
            - button_size: Size of the button.
            - loading: If True, the button will show loading animation.
            - disabled: If True, the button will be disabled.
            - icon: Icon to be displayed on the button.
            - available_in_offline: If True, the button will be available in offline session.
        """
        return {
            "check_existing_task_cb": self._check_existing_task_cb,
            "options": {
                "text": self._text,
                "button_type": self._button_type,
                "plain": self._plain,
                "button_size": self._button_size,
                "loading": self._loading,
                "disabled": self._disabled,
                "icon": self._icon,
                "available_in_offline": self._available_in_offline,
            },
        }

    def get_json_state(self) -> None:
        """Button widget doesn't have state, so this method returns None."""
        return {
            "team_id": self._team_id,
            "workspace_id": self._workspace_id,
            "module_id": self._module_id,
            "payload": self._payload,
            "visible_by_vue_field": self._visible_by_vue_field,
        }

    @property
    def workspace_id(self) -> int:
        """Returns the workspace ID.

        :returns: Workspace ID.
        :rtype: int
        """
        return self._workspace_id

    @workspace_id.setter
    def workspace_id(self, value: int) -> None:
        """Sets the workspace ID.

        :param value: Workspace ID.
        :type value: int
        """
        self._workspace_id = value
        StateJson()[self.widget_id]["workspace_id"] = self._workspace_id
        StateJson().send_changes()

    @property
    def text(self) -> str:
        """Returns the text to be displayed on the button.

        :returns: Text to be displayed on the button.
        :rtype: str
        """
        return self._text

    @text.setter
    def text(self, value: str) -> None:
        """Sets the text to be displayed on the button.

        :param value: Text to be displayed on the button.
        :type value: str
        """
        self._text = value
        DataJson()[self.widget_id]["options"]["text"] = self._text
        DataJson().send_changes()

    @property
    def payload(self) -> dict:
        """Returns the payload to be sent to the app.

        :returns: Payload to be sent to the app.
        :rtype: dict
        """
        return self._payload

    @payload.setter
    def payload(self, value: dict) -> None:
        """Sets the payload to be sent to the app.

        :param value: Payload to be sent to the app.
        :type value: dict
        """
        self._payload = value
        StateJson()[self.widget_id]["payload"] = self._payload
        StateJson().send_changes()

    @property
    def icon(self) -> str:
        """Returns the icon to be displayed on the button.

        :returns: Icon to be displayed on the button.
        :rtype: str
        """
        return self._icon

    @icon.setter
    def icon(self, value: str) -> None:
        """Sets the icon to be displayed on the button.

        :param value: Icon to be displayed on the button.
        :type value: str
        """
        if value is None:
            self._icon = ""
        else:
            self._icon = f'<i class="{value}" style="margin-right: {self._icon_gap}px"></i>'
        DataJson()[self.widget_id]["options"]["icon"] = self._icon
        DataJson().send_changes()

    @property
    def button_type(self) -> str:
        """Returns the type of the button.

        :returns: Type of the button.
        :rtype: str
        """
        return self._button_type

    @button_type.setter
    def button_type(
        self, value: Literal["primary", "info", "warning", "danger", "success", "text"]
    ) -> None:
        """Sets the type of the button.

        :param value: Type of the button.
        :type value: Literal["primary", "info", "warning", "danger", "success", "text"]
        """
        self._button_type = value
        DataJson()[self.widget_id]["options"]["button_type"] = self._button_type
        DataJson().send_changes()

    @property
    def plain(self) -> bool:
        """Returns True if the button is plain, False otherwise.

        :returns: True if the button is plain, False otherwise.
        :rtype: bool
        """
        return self._plain

    @plain.setter
    def plain(self, value: bool) -> None:
        """Sets the button to be plain or not.

        :param value: If True, the button will be plain.
        :type value: bool
        """
        self._plain = value
        DataJson()[self.widget_id]["options"]["plain"] = self._plain
        DataJson().send_changes()

    @property
    def loading(self) -> bool:
        """Returns True if the button shows loading animation, False otherwise.

        :returns: True if the button shows loading animation, False otherwise.
        :rtype: bool
        """
        return self._loading

    @loading.setter
    def loading(self, value: bool) -> None:
        """Sets the button loading animation.

        :param value: If True, the animation will be enabled, otherwise disabled.
        :type value: bool
        """
        self._loading = value
        DataJson()[self.widget_id]["options"]["loading"] = self._loading
        DataJson().send_changes()

    @property
    def disabled(self) -> bool:
        """Returns True if the button is disabled, False otherwise.

        :returns: True if the button is disabled, False otherwise.
        :rtype: bool
        """
        return self._disabled

    @disabled.setter
    def disabled(self, value: bool) -> None:
        """Sets the button to be disabled or not.

        :param value: If True, the button will be disabled.
        :type value: bool
        """
        self._disabled = value
        DataJson()[self.widget_id]["options"]["disabled"] = self._disabled

    def set_check_existing_task_cb(self, function: str) -> None:
        """Sets the callback function for checking existing tasks.
        Function should be a string (docstring) of JavaScript code.

        :param function: Callback function for checking existing tasks.
        :type function: str
        """
        self._check_existing_task_cb = function
        DataJson()[self.widget_id]["check_existing_task_cb"] = self._check_existing_task_cb
        DataJson().send_changes()



import json
import re
import textwrap

try:
    import markdown  # type: ignore
except ImportError:  # pragma: no cover
    markdown = None  # type: ignore
from jinja2.ext import Extension
from jinja2.nodes import CallBlock


class MarkdownExtension(Extension):
    """Jinja2 extension that renders a `{% markdown %}` block to HTML using Python-Markdown."""

    EXTENSIONS = [
        "admonition",
        "attr_list",
        "smarty",
        "tables",
        "pymdownx.betterem",
        "pymdownx.caret",
        "pymdownx.details",
        "pymdownx.emoji",
        "pymdownx.keys",
        "pymdownx.magiclink",
        "pymdownx.mark",
        "pymdownx.smartsymbols",
        "pymdownx.superfences",
        "pymdownx.tabbed",
        "pymdownx.tasklist",
        "pymdownx.tilde",
        "pymdownx.highlight",
    ]

    EXTENSION_CONFIGS = {
        "pymdownx.highlight": {
            "use_pygments": True,
            "noclasses": True,
            # "linenums": True,
            # "linenums_style": "pymdownx-inline",
        },
    }
    tags = set(["markdown"])

    def __init__(self, environment):
        """
        :param environment: Jinja2 Environment.
        :type environment: jinja2.Environment
        """
        super(MarkdownExtension, self).__init__(environment)
        if markdown is None:
            raise ImportError("markdown library is required for MarkdownExtension but is not installed.")

        environment.extend(
            markdowner=markdown.Markdown(
                extensions=self.EXTENSIONS, extension_configs=self.EXTENSION_CONFIGS
            )
        )

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        body = parser.parse_statements(["name:endmarkdown"], drop_needle=True)
        return CallBlock(self.call_method("_render_markdown"), [], [], body).set_lineno(lineno)

    def _render_markdown(self, caller):
        text = caller()
        text = self._dedent(text)
        return self.environment.markdowner.convert(text)

    def _dedent(self, text):
        return textwrap.dedent(text.strip("\n"))


class AutoSidebarExtension(Extension):
    """Jinja2 extension that injects an autosidebar placeholder into the rendered HTML."""

    tags = {"autosidebar"}

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        return CallBlock(self.call_method("_placeholder"), [], [], []).set_lineno(lineno)

    def _placeholder(self, caller):
        return "<!--AUTOSIDEBAR_PLACEHOLDER-->\n"


def _slugify(text: str) -> str:
    """Simple slugify helper (kebab-case)."""
    text = text.lower()
    text = re.sub(r"[^a-z0-9]+", "-", text)
    text = re.sub(r"-+", "-", text).strip("-")
    return text or "tab"


class TabExtension(Extension):
    """
    Internal helper extension. Each `{% tab title="X" %} ... {% endtab %}` inside a
    `{% tabs %} ... {% endtabs %}` block will register its content in a temporary
    buffer on the Jinja2 `Environment` (``_tab_buffer``).
    The buffer is consumed by the outer `TabsExtension` which renders the final
    HTML with `<sly-iw-tabs>` component.
    """

    tags = {"tab"}

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        # Expect: tab title="Some Title"
        parser.stream.expect("name:title")
        parser.stream.expect("assign")  # '=' symbol
        title_expr = parser.parse_expression()
        body = parser.parse_statements(["name:endtab"], drop_needle=True)
        return CallBlock(self.call_method("_render_tab", [title_expr]), [], [], body).set_lineno(
            lineno
        )

    def _render_tab(self, title: str, caller):
        raw_content = caller()
        # content_html = self.environment.markdowner.convert(raw_content)

        buf = getattr(self.environment, "_tab_buffer", None)
        if buf is not None:
            buf.append((title, raw_content))
            # buf.append((title, content_html))
        return ""


class TabsExtension(Extension):
    """
    Implements a GitBook-style tabs syntax:

    ```
    {% tabs %}
    {% tab title="Windows" %}
    Windows content
    {% endtab %}
    {% tab title="Linux" %}
    Linux content
    {% endtab %}
    {% endtabs %}
    ```
    """

    tags = {"tabs"}

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        body = parser.parse_statements(["name:endtabs"], drop_needle=True)
        return CallBlock(self.call_method("_render_tabs"), [], [], body).set_lineno(lineno)

    def _render_tabs(self, caller):
        # Prepare a buffer for nested tab blocks
        self.environment._tab_buffer = []
        caller()  # Render inner content – `TabExtension` will fill the buffer
        tabs_data = self.environment._tab_buffer
        # Clean up
        delattr(self.environment, "_tab_buffer")

        if not tabs_data:
            # Nothing collected – return original caller() content
            return ""

        # Build tab definitions
        tab_defs = []
        rendered_templates = []
        for title, content in tabs_data:
            slug = _slugify(title)
            tab_defs.append({"name": slug, "title": title})
            rendered_templates.append(f"<template #{slug}>\n{content}\n</template>")

        tabs_json = json.dumps(tab_defs)
        html_parts = [
            "<div>",
            f"<sly-iw-tabs :tabs='{tabs_json}' :defaultIndex='0' :command='command' :data='data'>",
            *rendered_templates,
            "</sly-iw-tabs>",
            "</div>",
        ]
        return "\n".join(html_parts)



from supervisely.app.widgets import Widget


class Identity(Widget):
    """Pass-through wrapper: renders child widget without adding layout or behavior. Useful for conditional display."""

    def __init__(self, content: Widget, widget_id: str = None):
        """
        :param content: Child widget to wrap.
        :type content: Widget
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._content = content
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {}

    def get_json_state(self):
        return {}



from typing import Dict, List, Tuple

import pandas as pd

from supervisely._utils import logger
from supervisely.api.api import Api
from supervisely.app.exceptions import show_dialog
from supervisely.app.widgets import Widget
from supervisely.app.widgets.container.container import Container
from supervisely.app.widgets.dropdown_checkbox_selector.dropdown_checkbox_selector import (
    DropdownCheckboxSelector,
)
from supervisely.app.widgets.fast_table.fast_table import FastTable


class EcosystemModelSelector(Widget):
    """Widget to browse and select models from the Supervisely Ecosystem with table filtering."""

    class COLUMN:
        """Column keys used in the internal models table."""

        MODEL_NAME = "name"
        FRAMEWORK = "framework"
        TASK_TYPE = "task"
        PARAMETERS = "params (M)"
        # TODO: support metrics for different tasks
        MAP = "mAP"

    COLUMNS = [
        str(COLUMN.MODEL_NAME),
        str(COLUMN.FRAMEWORK),
        str(COLUMN.TASK_TYPE),
        str(COLUMN.PARAMETERS),
        str(COLUMN.MAP),
    ]

    def __init__(self, frameworks: List[str] = None, task_types: List[str] = None, models: List[Dict] = None, api: Api = None, widget_id: str = None):
        """
        :param frameworks: Optional list of framework names to filter by.
        :type frameworks: List[str], optional
        :param task_types: Optional list of task types to filter by.
        :type task_types: List[str], optional
        :param models: Optional pre-loaded models list.
        :type models: List[Dict], optional
        :param api: Supervisely API. Defaults to Api().
        :type api: Api, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        if api is None:
            api = Api()
        self.api = api
        self.frameworks = None
        self.task_types = None
        self.models = None

        self.set(frameworks, task_types, models)

        self.framework_filter = DropdownCheckboxSelector(items=[], label="Frameworks")
        self.task_filter = DropdownCheckboxSelector(items=[], label="Task Types")
        self.table = FastTable(columns=self.COLUMNS, page_size=10, is_radio=True, header_right_content=Container(widgets=[self.framework_filter, self.task_filter], direction="horizontal"))
        self.refresh_table()
        self.table.set_filter(self._filter_function)

        @self.framework_filter.value_changed
        def _framework_filter_changed(frameworks: List[DropdownCheckboxSelector.Item]):
            self.frameworks = [f.id for f in frameworks] if frameworks else None
            task_types = [t.id for t in self.task_filter.get_selected()]
            self.task_types = task_types if task_types else None
            self.table.filter((self.frameworks, self.task_types))

        @self.task_filter.value_changed
        def _task_filter_changed(task_types: List[DropdownCheckboxSelector.Item]):
            self.task_types = [t.id for t in task_types] if task_types else None
            frameworks = [f.id for f in self.framework_filter.get_selected()]
            self.frameworks = frameworks if frameworks else None
            self.table.filter((self.frameworks, self.task_types))

        super().__init__(widget_id=widget_id, file_path=__file__)

    def _filter_function(
        self, data: pd.DataFrame, filter_value: Tuple[List[str], List[str]]
    ) -> pd.DataFrame:
        try:
            frameworks, task_types = filter_value
            if not frameworks and not task_types:
                return data
            filtered_models = set()
            for idx, model in enumerate(self.models):
                should_add = True
                if frameworks and model["framework"] not in frameworks:
                    should_add = False
                if task_types and model["task"] not in task_types:
                    should_add = False
                if should_add:
                    filtered_models.add(idx)

            filtered_data = data.iloc[sorted(filtered_models)]
            filtered_data.reset_index(inplace=True, drop=True)
        except Exception as e:
            logger.error(f"Error during filtering: {e}", exc_info=True)
            show_dialog(title="Filtering Error", description=str(e), status="error")
            return data
        return filtered_data

    def _filter_models(self, models: List[Dict], frameworks: List[str], task_types: List[str]) -> List[Dict]:
        if frameworks is None and task_types is None:
            return models

        filtered_models = []
        for model in models:
            if frameworks and model["framework"] not in frameworks:
                continue
            if task_types and model["task"] not in task_types:
                continue
            filtered_models.append(model)
        return filtered_models

    def set(self, frameworks: List[str] = None, task_types: List[str] = None, models: List[Dict] = None):
        self.frameworks = frameworks
        self.task_types = task_types
        if models is None:
            models = self.api.nn.ecosystem_models_api.list_models()
        self.models = models
        self.models = self._filter_models(self.models, self.frameworks, self.task_types)

    def _map_from_model(self, model: Dict):
        try:
            map = model.get("evaluation", {}).get("metrics", {}).get("mAP", None)
            if map is None:
                return None
            return float(map)
        except:
            return None

    def _params_from_model(self, model: Dict):
        try:
            params = model.get("paramsM", None)
            if params is None:
                return None
            return float(params)
        except:
            return None

    def _get_table_data(self, models: List[Dict] = None) -> List[Dict]:
        if models is None:
            models = self.models
        data = [
            {
                self.COLUMN.FRAMEWORK: model_data["framework"],
                self.COLUMN.MODEL_NAME: model_data["name"],
                self.COLUMN.TASK_TYPE: model_data["task"],
                self.COLUMN.PARAMETERS: self._params_from_model(model_data),
                self.COLUMN.MAP: self._map_from_model(model_data),
            }
            for model_data in models
        ]
        return data

    def refresh_table(self):
        data = self._get_table_data()
        df = pd.DataFrame.from_records(data=data, columns=self.COLUMNS)
        self.table.read_pandas(df)
        unique_frameworks = df[self.COLUMN.FRAMEWORK].unique().tolist()
        self.framework_filter.set(items=[DropdownCheckboxSelector.Item(id=f, name=f) for f in unique_frameworks])
        unique_task_types = df[self.COLUMN.TASK_TYPE].unique().tolist()
        self.task_filter.set(items=[DropdownCheckboxSelector.Item(id=t, name=t) for t in unique_task_types])

    def get_selected(self):
        idx = self.table.get_selected_row().row_index
        models = self._filter_models(self.models, self.frameworks, self.task_types)
        return models[idx]

    def select(self, index: int):
        self.table.select_row(index)

    def select_framework_and_model_name(self, framework: str, model_name: str):
        for idx, model in enumerate(self.models):
            if model["name"] == model_name and model["framework"] == framework:
                self.table.select_row(idx)
                return
        raise ValueError(f"Model with framework `{framework}` and name '{model_name}' not found.")

    def get_json_data(self):
        return {}

    def get_json_state(self):
        return {}

    def to_html(self):
        return self.table.to_html()

    def disable(self):
        return self.table.disable()

    def enable(self):
        return self.table.enable()

    def hide(self):
        return self.table.hide()

    def show(self):
        return self.table.show()

    @property
    def loading(self):
        return self.table.loading

    @loading.setter
    def loading(self, value: bool):
        self.table.loading = value



from __future__ import annotations

import uuid
from typing import Dict, Optional, Tuple, Union

from supervisely._utils import take_with_default
from supervisely.annotation.tag_meta import TagMeta
from supervisely.annotation.tag_meta_collection import TagMetaCollection
from supervisely.api.module_api import ApiField
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_tag import VideoTag


class PointcloudEpisodeTag(VideoTag):
    """Tag on point cloud episode or frame range (meta, value, frame_range). Immutable."""

    _SUPPORT_UNFINISHED_TAGS = True

    def __init__(
        self,
        meta,
        value=None,
        frame_range=None,
        key=None,
        sly_id=None,
        labeler_login=None,
        updated_at=None,
        created_at=None,
        is_finished=None,
        non_final_value=None,
    ):
        """
        Tag on point cloud episode or frame range.

        :param meta: Tag metadata (name, value type).
        :type meta: :class:`~supervisely.annotation.tag_meta.TagMeta`
        :param value: Tag value; type must match TagMeta.value_type.
        :type value: str or int or float, optional
        :param frame_range: Frame range (start, end) where tag applies.
        :type frame_range: Tuple[int, int] or List[int, int], optional
        :param key: UUID key. Auto-generated if not provided.
        :type key: uuid.UUID, optional
        :param sly_id: Server-side tag ID.
        :type sly_id: int, optional
        :param labeler_login: Login of user who created the tag.
        :type labeler_login: str, optional
        :param updated_at: Last modification timestamp (ISO format).
        :type updated_at: str, optional
        :param created_at: Creation timestamp (ISO format).
        :type created_at: str, optional
        :param is_finished: Whether range tag is finalized.
        :type is_finished: bool, optional
        :param non_final_value: Whether tag value is temporary.
        :type non_final_value: bool, optional

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                meta_car = sly.TagMeta('car', sly.TagValueType.NONE)
                tag_car = sly.PointcloudEpisodeTag(meta_car)

                meta_cat = sly.TagMeta('cat', sly.TagValueType.ANY_STRING)
                tag_cat = sly.PointcloudEpisodeTag(meta_cat, value="red", frame_range=(5, 10))
        """
        super().__init__(
            meta,
            value,
            frame_range,
            key,
            sly_id,
            labeler_login,
            updated_at,
            created_at,
            is_finished=is_finished,
            non_final_value=non_final_value,
        )

    @classmethod
    def from_json(
        cls,
        data: Dict,
        tag_meta_collection: TagMetaCollection,
        key_id_map: Optional[KeyIdMap] = None,
    ) -> PointcloudEpisodeTag:
        """
        Convert a json dict to PointcloudEpisodeTag. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: PointcloudEpisodeTag in json format as a dict.
        :type data: dict
        :param tag_meta_collection: Tag metadata collection.
        :type tag_meta_collection: :class:`~supervisely.annotation.tag_meta_collection.TagMetaCollection`
        :param key_id_map: Key ID map.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`
        :returns: Pointcloud episode tag object.
        :rtype: :class:`~supervisely.pointcloud_annotation.pointcloud_episode_tag.PointcloudEpisodeTag`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                tag_car_color_json = {
                    "frameRange": [15, 20],
                    "key": "da9ca75e97744fc5aaf24d6be2eb2832",
                    "name": "car color",
                    "value": "white"
                }

                colors = ["brown", "white", "black", "red", "chocolate", "gold", "grey"]
                meta_car_color = sly.TagMeta('car color', sly.TagValueType.ONEOF_STRING, possible_values=colors)
                meta_car_collection = sly.TagMetaCollection([meta_car_color])

                tag_car_color = sly.PointcloudEpisodeTag.from_json(tag_car_color_json, meta_car_collection)
        """
        is_finished = data.get(ApiField.IS_FINISHED, True)
        non_final_value = data.get(ApiField.NON_FINAL_VALUE, False)
        temp = super(PointcloudEpisodeTag, cls).from_json(data, tag_meta_collection, key_id_map)

        return cls(
            meta=temp.meta,
            value=temp.value,
            frame_range=temp.frame_range,
            key=temp.key(),
            sly_id=temp.sly_id,
            labeler_login=temp.labeler_login,
            updated_at=temp.updated_at,
            created_at=temp.created_at,
            is_finished=is_finished,
            non_final_value=non_final_value,
        )

    def __eq__(self, other: PointcloudEpisodeTag) -> bool:
        return super().__eq__(other)

    def clone(
        self,
        meta: Optional[TagMeta] = None,
        value: Optional[Union[str, int, float]] = None,
        frame_range: Optional[Tuple[int, int]] = None,
        key: Optional[uuid.UUID] = None,
        sly_id: Optional[int] = None,
        labeler_login: Optional[str] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
        is_finished: Optional[bool] = None,
        non_final_value: Optional[bool] = None,
    ) -> PointcloudEpisodeTag:
        """
        Makes a copy of PointcloudEpisodeTag with new fields, if fields are given, otherwise it will use fields of the original PointcloudEpisodeTag.

        :param meta: General information about pointcloud episode tag.
        :type meta: :class:`~supervisely.annotation.tag_meta.TagMeta`, optional
        :param value: PointcloudEpisodeTag value. Depends on TagValueType of :class:`~supervisely.annotation.tag_meta.TagMeta`.
        :type value: Optional[Union[str, int, float]]
        :param frame_range: PointcloudEpisodeTag frame range.
        :type frame_range: Optional[Union[Tuple[int, int], List[int, int]]]
        :param key: uuid.UUID object.
        :type key: uuid.UUID, optional
        :param sly_id: PointcloudEpisodeTag ID in Supervisely.
        :type sly_id: int, optional
        :param labeler_login: Login of user who created :class:`~supervisely.pointcloud_annotation.pointcloud_episode_tag.PointcloudEpisodeTag`.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when PointcloudEpisodeTag was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when PointcloudEpisodeTag was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional
        :param is_finished: Pointcloud Episode Tag is finished or not (applicable for range tags).
        :type is_finished: bool, optional
        :param non_final_value: Pointcloud Episode Tag value is final or not. Can be useful to create tag without value.
        :type non_final_value: bool, optional

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                colors = ["brown", "white", "black", "red", "chocolate", "gold", "grey"]
                meta_car_color = sly.TagMeta('car color', sly.TagValueType.ONEOF_STRING, possible_values=colors)
                tag_car_color = sly.PointcloudEpisodeTag(meta_car_color, value="white", frame_range=(15, 20))
                meta_bus = sly.TagMeta('bus', sly.TagValueType.ANY_STRING)

                new_tag = tag_car_color.clone(meta=meta_bus, frame_range=(15, 30), key=tag_car_color.key())
                print(new_tag.to_json())
                # Output: {
                #     "frameRange": [15, 30],
                #     "key": "4360b25778144141aa4f1a0d775a0a7a",
                #     "name": "bus",
                #     "value": "white"
                # }
        """

        return self.__class__(
            meta=take_with_default(meta, self.meta),
            value=take_with_default(value, self.value),
            frame_range=take_with_default(frame_range, self.frame_range),
            key=take_with_default(key, self.key()),
            sly_id=take_with_default(sly_id, self.sly_id),
            labeler_login=take_with_default(labeler_login, self.labeler_login),
            updated_at=take_with_default(updated_at, self.updated_at),
            created_at=take_with_default(created_at, self.created_at),
            is_finished=take_with_default(is_finished, self.is_finished),
            non_final_value=take_with_default(non_final_value, self.non_final_value),
        )



import time


class TinyTimer:
    """Lightweight timer for measuring elapsed time since creation."""

    def __init__(self):
        self.t = time.time()

    def get_sec(self):  # since creation
        now_t = time.time()
        return now_t - self.t


from typing import List, Tuple

from supervisely.annotation.annotation import Annotation
from supervisely.api.image_api import ImageInfo
from supervisely.api.module_api import ApiField
from supervisely.sly_logger import logger
from supervisely.nn.benchmark.base_visualizer import BaseVisMetrics
from supervisely.nn.benchmark.visualization.widgets import GalleryWidget, MarkdownWidget
from supervisely.project.project_meta import ProjectMeta


class ExplorePredictions(BaseVisMetrics):
    """Gallery and widgets to explore prediction differences across compared detection models."""

    MARKDOWN_DIFFERENCE = "markdown_explore_difference"
    GALLERY_DIFFERENCE = "explore_difference_gallery"
    MARKDOWN_SAME_ERRORS = "markdown_explore_same_errors"
    GALLERY_SAME_ERRORS = "explore_same_error_gallery"

    @property
    def difference_predictions_md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_explore_difference
        return MarkdownWidget(self.MARKDOWN_DIFFERENCE, "Explore Predictions", text)

    @property
    def explore_gallery(self) -> GalleryWidget:
        columns_number = len(self.eval_results) + 1
        *data, min_conf = self._get_sample_data()
        default_filters = [{"confidence": [min_conf, 1]}]
        gallery = GalleryWidget(
            self.GALLERY_DIFFERENCE,
            columns_number=columns_number,
            filters=default_filters,
        )
        gallery.add_image_left_header("Click to explore more")
        gallery.set_project_meta(self.eval_results[0].gt_project_meta)
        gallery.set_images(*data)
        click_data = self.get_click_data_explore_all()
        gallery.set_click_data(self.explore_modal_table.id, click_data)
        gallery.set_show_all_data(self.explore_modal_table.id, click_data)
        gallery._gallery._update_filters()

        return gallery

    def _get_sample_data(self) -> Tuple[List[ImageInfo], List[Annotation], List[ProjectMeta]]:
        images = []
        annotations = []
        metas = [self.eval_results[0].gt_project_meta]
        skip_tags_filtering = [True]
        min_conf = float("inf")
        sample_keys = self._get_evaluated_image_keys(limit=5)
        if len(sample_keys) == 0:
            raise RuntimeError("Failed to prepare sample gallery: no evaluated images were found.")

        gt_eval_res = self.eval_results[0]
        gt_map = self._get_project_items_by_keys(gt_eval_res.gt_project_id, sample_keys)
        column_maps = [gt_map]

        for eval_res in self.eval_results:
            metas.append(eval_res.pred_project_meta)
            pred_map = self._get_project_items_by_keys(eval_res.pred_project_id, sample_keys)
            column_maps.append(pred_map)
            skip_tags_filtering.append(False)
            min_conf = min(min_conf, eval_res.mp.conf_threshold)

        common_keys = [key for key in sample_keys if all(key in col_map for col_map in column_maps)]
        if len(common_keys) == 0:
            raise RuntimeError(
                "Failed to prepare sample gallery: no common evaluated images were found "
                "across GT and prediction projects."
            )

        for key in common_keys:
            for col_map in column_maps:
                image_info, ann_info = col_map[key]
                images.append(image_info)
                annotations.append(ann_info)
        return images, annotations, metas, skip_tags_filtering, min_conf

    def _get_evaluated_image_keys(self, limit: int = None) -> List[Tuple[str, str]]:
        """Returns evaluated image keys as (dataset_name, image_name) pairs"""
        keys = sorted(
            {
                (img.get("dataset"), img.get("file_name"))
                for img in self.eval_results[0].coco_gt.imgs.values()
                if img.get("dataset") is not None and img.get("file_name") is not None
            }
        )
        if limit is not None:
            return keys[:limit]
        return keys

    def _get_dataset_info(self, project_id: int, dataset_name: str):
        api = self.eval_results[0].api
        ds_info = api.dataset.get_info_by_name(project_id, dataset_name)
        if ds_info is not None:
            return ds_info
        for ds in api.dataset.get_list(project_id, recursive=True):
            if ds.name == dataset_name:
                return ds
        return None

    def _get_project_items_by_keys(self, project_id: int, keys: List[Tuple[str, str]]):
        """
        Build mapping (dataset_name, image_name) -> (ImageInfo, AnnotationInfo) for given project.
        """
        api = self.eval_results[0].api
        result = {}

        ds_to_names = {}
        for ds_name, img_name in keys:
            ds_to_names.setdefault(ds_name, []).append(img_name)

        for ds_name, img_names in ds_to_names.items():
            if len(img_names) == 0:
                continue
            ds_info = self._get_dataset_info(project_id, ds_name)
            if ds_info is None:
                logger.warning(
                    f"Dataset '{ds_name}' not found in project {project_id}. Skipping.",
                    extra={"project_id": project_id, "dataset_name": ds_name},
                )
                continue

            infos = api.image.get_list(
                ds_info.id,
                filters=[
                    {ApiField.FIELD: ApiField.NAME, ApiField.OPERATOR: "in", ApiField.VALUE: img_names}
                ],
                force_metadata_for_links=False,
            )
            if len(infos) == 0:
                continue

            infos = sorted(infos, key=lambda x: img_names.index(x.name))
            image_ids = [info.id for info in infos]
            ann_infos = api.annotation.download_batch(
                ds_info.id, image_ids, force_metadata_for_links=False
            )
            ann_infos_by_id = {ann.image_id: ann for ann in ann_infos}

            for info in infos:
                ann = ann_infos_by_id.get(info.id)
                if ann is None:
                    continue
                key = (ds_name, info.name)
                if key not in result:
                    result[key] = (info, ann)
        return result

    def get_click_data_explore_all(self) -> dict:
        res = {}

        res["projectMeta"] = self.eval_results[0].gt_project_meta.to_json()
        res["layoutTemplate"] = [None, None, None]

        res["layoutTemplate"] = [{"skipObjectTagsFiltering": True, "columnTitle": "Ground Truth"}]
        # for i in range(len(self.eval_results)):
        for idx, eval_res in enumerate(self.eval_results, 1):
            res["layoutTemplate"].append({"columnTitle": f"[{idx}] {eval_res.name}"})

        click_data = res.setdefault("clickData", {})
        explore = click_data.setdefault("explore", {})
        explore["title"] = "Explore all predictions"

        images_ids = []
        min_conf = float("inf")
        all_keys = self._get_evaluated_image_keys()

        gt_eval_res = self.eval_results[0]
        gt_map = self._get_project_items_by_keys(gt_eval_res.gt_project_id, all_keys)
        pred_maps = []
        for eval_res in self.eval_results:
            pred_maps.append(self._get_project_items_by_keys(eval_res.pred_project_id, all_keys))
            min_conf = min(min_conf, eval_res.mp.conf_threshold)

        column_maps = [gt_map] + pred_maps
        common_keys = [key for key in all_keys if all(key in col_map for col_map in column_maps)]
        if len(common_keys) == 0:
            raise RuntimeError(
                "Failed to build explore data: no common evaluated images were found "
                "across GT and prediction projects."
            )

        for col_map in column_maps:
            images_ids.append([col_map[key][0].id for key in common_keys])

        explore["imagesIds"] = list(i for x in zip(*images_ids) for i in x)
        explore["filters"] = [{"type": "tag", "tagId": "confidence", "value": [min_conf, 1]}]

        return res



from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional, Union

from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class Carousel(Widget):
    """Cycles through a series of images or text in limited space."""

    class Routes:
        """Callback route names used by the widget frontend to notify Python."""
        VALUE_CHANGED = "value_changed"

    class Item:
        """Single carousel item (name, label, is_link)."""

        def __init__(
            self,
            name: Optional[str] = "",
            label: Optional[str] = "",
            is_link: Optional[bool] = True,
        ):
            """
            :param name: Item name.
            :param label: Item label.
            :param is_link: Display as link.
            """
            self.name = name
            self.label = label
            self.is_link = is_link

        def to_json(self) -> Dict[str, Union[str, bool]]:
            """Returns dictionary with item data.

            Dictionary contains the following fields:
                - name: Name of the item.
                - label: Label of the item.
                - is_link: If True, the item will be displayed as a link.
            """
            return {
                "name": self.name,
                "label": self.label,
                "is_link": self.is_link,
            }

    def __init__(
        self,
        items: List[Carousel.Item],
        height: Optional[int] = 350,
        initial_index: Optional[int] = 0,
        trigger: Optional[Literal["hover", "click"]] = "click",
        autoplay: Optional[bool] = False,
        interval: Optional[int] = 3000,
        indicator_position: Optional[Literal["outside", "none"]] = "none",
        arrow: Optional[Literal["always", "hover", "never"]] = "hover",
        type: Optional[Literal["card"]] = None,
        widget_id: Optional[str] = None,
    ):
        """
        :param items: List of Carousel.Item.
        :param height: Height in pixels.
        :param initial_index: Initial visible item index.
        :param trigger: "hover" or "click".
        :param autoplay: Enable autoplay.
        :param interval: Autoplay interval (ms).
        :param indicator_position: "outside" or "none".
        :param arrow: "always", "hover", or "never".
        :param type: "card" or None.
        :param widget_id: Widget identifier.

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import Carousel
                items = [Carousel.Item(name="i1", label="Item 1"), Carousel.Item(name="i2", label="Item 2")]
                carousel = Carousel(items=items, height=350, trigger="click")
        """
        self._height = f"{height}px"
        self._items = items
        self._initial_index = initial_index
        self._trigger = trigger
        self._autoplay = autoplay
        self._interval = interval
        self._indicator_position = indicator_position if indicator_position is not None else "none"
        self._arrow = arrow
        self._type = type

        self._changes_handled = False
        self._clicked_value = None

        super().__init__(widget_id=widget_id, file_path=__file__)

    def _set_items(self):
        return [item.to_json() for item in self._items]

    def get_json_data(self) -> Dict[str, Any]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - height: Height of the carousel in pixels.
            - items: List of items to be displayed in the carousel.
            - initialIndex: Index of the item to be displayed at initialization.
            - trigger: Trigger type to change the active item.
            - autoplay: If True, the carousel will be autoplayed.
            - interval: Time interval between each autoplay.
            - indicatorPosition: Position of the indicator.
            - arrow: Arrow display type.
            - type: Carousel type.
        """

        return {
            "height": self._height,
            "items": self._set_items(),
            "initialIndex": self._initial_index,
            "trigger": self._trigger,
            "autoplay": self._autoplay,
            "interval": self._interval,
            "indicatorPosition": self._indicator_position,
            "arrow": self._arrow,
            "type": self._type,
        }

    def get_json_state(self) -> Dict[str, Any]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
            - clickedValue: Name of the item that was clicked.

        :returns: Dictionary with widget state.
        :rtype: Dict[str, Any]
        """
        return {"clickedValue": self._clicked_value}

    def get_active_item(self) -> int:
        """Returns index of the active item.

        :returns: Index of the active item.
        :rtype: int
        """
        return StateJson()[self.widget_id]["clickedValue"]

    def get_items(self) -> List[Carousel.Item]:
        """Returns list of items.

        :returns: List of items.
        :rtype: List[:class:`~supervisely.app.widgets.carousel.carousel.Carousel.Item`]
        """
        return DataJson()[self.widget_id]["items"]

    def set_items(self, value: List[Carousel.Item]) -> None:
        """Sets list of items to be displayed in the carousel.
        This method will overwrite the existing items, not append to it.
        To append items, use :meth:`add_items`.

        :param value: List of items to be displayed in the carousel.
        :type value: List[:class:`~supervisely.app.widgets.carousel.carousel.Carousel.Item`]
        :raises ValueError: If items are not of type :class:`~supervisely.app.widgets.carousel.carousel.Carousel.Item`.
        """
        if not all(isinstance(item, Carousel.Item) for item in value):
            raise ValueError("Items must be of type Carousel.Item")
        self._items = value
        DataJson()[self.widget_id]["items"] = self._set_items()
        DataJson().send_changes()

    def add_items(self, value: List[Carousel.Item]) -> None:
        """Appends list of items to the existing items.
        This method will not overwrite the existing items, but append to it.
        To overwrite items, use :meth:`set_items`.

        :param value: List of items to be displayed in the carousel.
        :type value: List[:class:`~supervisely.app.widgets.carousel.carousel.Carousel.Item`]
        """
        self._items.extend(value)
        DataJson()[self.widget_id]["items"] = self._set_items()
        DataJson().send_changes()

    def set_height(self, value: int) -> None:
        """Sets height of the carousel.

        :param value: Height of the carousel in pixels.
        :type value: int
        :raises ValueError: If height value is not an integer.
        """
        if type(value) is not int:
            raise ValueError("Height value must be an integer")
        self._height = f"{value}px"
        DataJson()[self.widget_id]["height"] = self._height
        DataJson().send_changes()

    def get_initial_index(self) -> int:
        """Returns index of the item to be displayed at initialization.

        :returns: Index of the item to be displayed at initialization.
        :rtype: int
        """
        return DataJson()[self.widget_id]["initialIndex"]

    def set_initial_index(self, value: int) -> None:
        """Sets index of the item to be displayed at initialization.

        :param value: Index of the item to be displayed at initialization.
        :type value: int
        :raises ValueError: If index value exceeds the size of the carousel items.
        """
        if value >= len(self._items):
            raise ValueError("Index of the value being set exceeds the size of the carousel items.")
        self._initial_index = value
        DataJson()[self.widget_id]["initialIndex"] = self._initial_index
        DataJson().send_changes()

    def value_changed(self, func: Callable[[str], Any]) -> Callable[[], None]:
        """Decorator for the function to be called when the value of the carousel is changed.

        :param func: Function to be called when the value of the carousel is changed.
        :type func: Callable[[str], Any]
        :returns: Decorated function.
        :rtype: Callable[[], None]
        """
        route_path = self.get_route_path(Carousel.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            curr_idx = self.get_active_item()
            curr_name = self._items[curr_idx].name
            func(curr_name)

        return _click



from __future__ import annotations
from collections import namedtuple
from supervisely.app.widgets import Widget
from supervisely.app.content import DataJson, StateJson

from typing import List, Dict, Union, Optional, Callable

class Transfer(Widget):
    """Widget for transferring items between source and target lists. Operates with a single list; use transferred_items for keys to show in target."""

    class Routes:
        """Route name constants for this widget."""
        VALUE_CHANGED = "value_changed"

    class Item:
        """Class for representing items in the Transfer widget."""

        def __init__(self, key: str, label: Optional[str] = None, disabled: Optional[bool] = False):
            """
            :param key: Unique key for identifying the item.
            :type key: str
            :param label: Label displayed in the widget. If None, key is used.
            :type label: Optional[str]
            :param disabled: If True, the item won't be transferable.
            :type disabled: Optional[bool]

            :Usage Example:

                .. code-block:: python

                    Transfer.Item(key="dog", label="Dog", disabled=True)
                    Transfer.Item(key="cat", label="Cat")
                    Transfer.Item(key="mouse")
            """
            self.key = key
            if not label:
                # If label is not specified, the key will be used as label.
                self.label = key
            else:
                self.label = label
            self.disabled = disabled

        def to_json(self):
            return {"key": self.key, "label": self.label, "disabled": self.disabled}

    def __init__(
        self,
        items: Optional[Union[List[Item], List[str]]] = None,
        transferred_items: Optional[List[str]] = None,
        widget_id: Optional[str] = None,
        filterable: Optional[bool] = False,
        filter_placeholder: Optional[str] = None,
        titles: Optional[List[str]] = None,
        button_texts: Optional[List[str]] = None,
        left_checked: Optional[List[str]] = None,
        right_checked: Optional[List[str]] = None,
        width: int = 150,
    ):
        """Initialize the Transfer widget.

        :param items: List of items or list of strings (keys, auto-transformed into items).
        :type items: Union[List[:class:`~supervisely.app.widgets.transfer.transfer.Transfer.Item`], List[str]], optional
        :param transferred_items: Keys of items to display in the right (target) list.
        :type transferred_items: List[str], optional
        :param widget_id: The id of the widget.
        :type widget_id: str, optional
        :param filterable: If True, the widget will have a filter input.
        :type filterable: bool, optional
        :param filter_placeholder: Placeholder for the filter input (only if filterable=True).
        :type filter_placeholder: str, optional
        :param titles: Titles for source and target lists. Default: ('Source', 'Target').
        :type titles: List[str], optional
        :param button_texts: Texts for transfer buttons. Default: icons only.
        :type button_texts: List[str], optional
        :param left_checked: Keys of items to check in left list at init.
        :type left_checked: List[str], optional
        :param right_checked: Keys of items to check in right list at init.
        :type right_checked: List[str], optional
        :param width: Width of the widget in pixels. Minimum 150.
        :type width: int, optional

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import Transfer

                item1 = Transfer.Item(key="cat", label="Cat", disabled=True)
                item2 = Transfer.Item(key="dog", label="Dog")
                transfer = Transfer(items=[item1, item2], transferred_items=["dog"])
                transfer.set_transferred_items(["cat"])
        """
        self._changes_handled = False
        self._items = []
        self._transferred_items = []

        if items:
            self._items = self.__checked_items(items)

        if transferred_items:
            self._transferred_items = self.__checked_transferred_items(transferred_items)

        # If wrong items are specified, items won't be checked.
        self._left_checked = left_checked
        self._right_checked = right_checked

        self._filterable = filterable
        self._filter_placeholder = filter_placeholder

        self._width = max(width, 150)

        self._titles = titles if titles is not None else ["Source", "Target"]

        self._button_texts = button_texts

        super().__init__(widget_id=widget_id, file_path=__file__)

    def __checked_items(self, items: Optional[Union[List[Item], List[str]]]) -> List[Transfer.Item]:
        """
        If the list of items is specified as a list of strings, they will be converted to items. List of
        Transfer items will be checked for uniqueness of the keys. If the keys of the items are not unique, an error will be
        raised.

        :param items: The list of items can either be a list of items or a list of strings, containing the
            keys for items to be created.
        :type items: Optional[Union[List[Item], List[str]]]

        :raises ValueError: If the keys of the items are not unique.

        :returns: The list of items with unique keys.
        :rtype: List[:class:`~supervisely.app.widgets.transfer.transfer.Transfer.Item`]
        """

        if isinstance(items[0], str):
            # If items are specified as strings, they will be converted to Transfer.Item objects.
            if len(set(items)) != len(items):
                # If the keys of the items are not unique, an error will be raised.
                raise ValueError("The keys of the items should be unique.")

            checked_items = [Transfer.Item(key=item) for item in items]
        else:
            # If items are specified as Transfer.Item objects, they will be checked for uniqueness.
            if len({item.key for item in items}) != len(items):
                # If the keys of the items are not unique, an error will be raised.
                raise ValueError("The keys of the items should be unique.")

            checked_items = items
        return checked_items

    def __checked_transferred_items(self, transferred_items: List[str]) -> List[str]:
        """
        If the self._items is specified, the list of transferred items will be checked for the keys of the items. Since
        transferred items are specified by keys of the items, each key of the transferred items should exist in the list of
        items. Otherwise, an error will be raised.

        :param transferred_items: List of keys of the items to be shown in the right (target) list.
        :type transferred_items: List[str]

        :raises ValueError: If transferred items are specified, but the list of items is not specified.
        :raises ValueError: If any of transferred items keys is not in the list of items.

        :returns: List of transferred items (keys of the items which should be displayed in the right list).
        :rtype: List[str]
        """

        if not self._items:
            # If the list of items is not specified, the list of transferred items can't be specified since
            # the keys of the items should be specified in the list of transferred items.
            raise ValueError("The 'items' argument should be specified if "\
                "the 'transferred_items' argument is specified.")
        else:
            if not set(transferred_items).issubset([item.key for item in self._items]):
                # If any of the keys of the items specified in the list of transferred items is not
                # in the list of items, an error will be raised.
                raise ValueError("The 'transferred_items' argument should contain only "\
                    "the keys of the items specified in the 'items' argument.")
            else:
                return transferred_items

    def get_json_data(self) -> Dict[str, Union[List[Dict[str, Union[str, bool]]], None]]:
        """
        Returns the data of the widget in JSON format.

        Data will contain the list of items and the list of transferred items.

        :returns: The data of the widget in JSON format: {"items": List[Dict[str, Any]], "transferred_items": List[str]}.
            "items" - the list of items in the widget in JSON format. Each item is represented as item.
            "transferred_items" - the list of transferred items (keys of the items which should be displayed in the right list).
        :rtype: Dict[str, List[Dict[str, Union[str, bool]]]]
        """

        res = {
            "items": [],
        }
        if self._items:
            res["items"] = [item.to_json() for item in self._items]

        return res

    def get_json_state(self) -> Dict[str, List[str]]:
        """
        Returns the state of the widget in JSON format.

        State will contain the list of transferred items.

        :returns: The state of the widget in JSON format: {"transferred_items": List[str]}. "transferred_items" - the list of
            transferred items (keys of the items which should be displayed in the right list).
        :rtype: Dict[str, List[str]]
        """

        transferred_items = self._transferred_items

        return {"transferred_items": transferred_items}

    def get_transferred_items(self) -> List[str]:        
        """
        Returns the list of transferred items.

        :returns: List of transferred items (keys of the items which should be displayed in the right list).
        :rtype: List[str]
        """

        return StateJson()[self.widget_id]["transferred_items"]

    def get_untransferred_items(self) -> List[str]:
        """
        Returns the list of untransferred items.

        :returns: List of untransferred items (keys of the items which should be displayed in the left list).
        :rtype: List[str]
        """

        return [item.key for item in self._items if item.key not in self.get_transferred_items()]

    def value_changed(self, func: Callable) -> Callable:
        """
        Decorates a function which will be called when the the items in right list are changed (moved in or out of the list).

        :param func: Function to be wrapped with the decorator. The function should have one argument which will contain
            namedtuple with the following fields: transferred_items, untransferred_items.
        :type func: Callable

        :returns: Wrapped function.
        :rtype: Callable

        :Usage Example:

            .. code-block:: python

                tr = Transfer(items=["item1", "item2", "item3"], transferred_items=["item1"])

                # Move "item2" from the left list to the right list.
                @tr.value_changed
                def func(items):
                    print(items.transferred_items) # ["item1", "item2"]
                    print(items.untransferred_items) # ["item3"]
        """

        route_path = self.get_route_path(Transfer.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():

            Items = namedtuple("Items", ["transferred_items", "untransferred_items"])

            res = Items(transferred_items=self.get_transferred_items(), untransferred_items=self.get_untransferred_items())

            func(res)

        return _click

    def set_items(self, items: Union[List[Transfer.Item], List[str]]):
        """
        Sets the list of items for the widget.

        If the list of items is specified as strings, they will be converted to items.
        Note: this method will REPLACE the current list of items with the new one. If you want to add new items to the current
        list, use .add() method.

        :param items: List of items to be set as the new list of items for the widget. Can be specified as items
            objects or as strings of the item keys.
        :type items: Union[List[:class:`~supervisely.app.widgets.transfer.transfer.Transfer.Item`], List[str]]

        :Usage Example:

            .. code-block:: python

                tr = Transfer(items=["cat", "dog"])
                print(tr.get_untransferred_items()) # ["cat", "dog"]

                tr.set(items=["bird", "mouse"])
                print(tr.get_untransferred_items()) # ["bird", "mouse"]

                # As you can see, the list of items was replaced with the new one.
        """

        if items:
            self._items = self.__checked_items(items)
        else:
            self._items = []

        self.update_data()
        DataJson().send_changes()

    def set_transferred_items(self, transferred_items: List[str]):
        """
        Sets the list of transferred items.
        The list should contain only the keys of the items specified in the list of items. Otherwise, an error will be raised.
        :param transferred_items: List of keys of the items which should be displayed in the right list.
        :type transferred_items: List[str]
        """

        self._transferred_items = self.__checked_transferred_items(transferred_items)
        self.update_state()
        StateJson().send_changes()

    def add(self, items: Union[List[Item], List[str]]):
        """
        Adds new items to the current list of items.

        If the list of items is specified as strings, items will be created from them. If the list of adding
        items contains any items with the same key as the items in the current list, an error will be raised.

        :param items: List of items to be added to the current list of items. Can be specified as items or
            as strings of the item keys.
        :type items: Union[List[:class:`~supervisely.app.widgets.transfer.transfer.Transfer.Item`], List[str]]

        :raises ValueError: If the list of adding items contains any items with the same key as the items in the current list.

        :Usage Example:

            .. code-block:: python

                tr = Transfer(items=["cat", "dog"])
                print(tr.get_untransferred_items()) # ["cat", "dog"]

                tr.add(items=["bird", "mouse"])
                print(tr.get_untransferred_items()) # ["cat", "dog", "bird", "mouse"]
        """

        items = self.__checked_items(items)

        if any([item.key in [item.key for item in self._items] for item in items]):
            raise ValueError("The 'items' argument should not contain any items with the same key as the items in the current list.")
        else:
            self._items.extend(items)
            self.update_data()
            DataJson().send_changes()

    def remove(self, items_keys: List[str]):
        """
        Removes items from the current list of items.

        The list of items to be removed should contain keys of the items which should be removed. If there are no items with 
        the specified keys in the current list, nothing will be removed and no error will be raised.

        :param items_keys: List of keys of the items to be removed.
        :type items_keys: List[str]
        """

        self._items = [item for item in self._items if item.key not in items_keys]
        self._transferred_items = [item for item in self._transferred_items if item not in items_keys]
        self.update_data()
        self.update_state()
        DataJson().send_changes()
        StateJson().send_changes()

    def get_items_keys(self) -> List[str]:
        """
        Returns the list of keys of the items.

        :returns: List of keys of the items.
        :rtype: List[str]
        """

        return [item.key for item in self._items]



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from typing import Any, Dict, Optional

import supervisely.io.env as sly_env
from supervisely.app import StateJson, DataJson
from supervisely.app.widgets import Widget
from supervisely.api.api import Api
from supervisely.nn.inference import Session


class ModelInfo(Widget):
    """Shows connected model details (name, checkpoint, task type) for a given session or team."""

    def __init__(
        self,
        session_id: int = None,
        team_id: int = None,
        widget_id: str = None,
        replace_none_with: Optional[str] = None,
    ):
        """:param session_id: Deployed model session ID to display info for.
        :type session_id: int, optional
        :param team_id: Team ID. Defaults to env.team_id().
        :type team_id: int, optional
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        :param replace_none_with: String to replace None values in model info.
        :type replace_none_with: str, optional
        """
        self._api = Api()
        self._session_id = session_id
        self._team_id = team_id
        self._model_info = None
        # used only if session data recieved from Session instance directly
        self._replace_none_with = replace_none_with

        if self._team_id is None:
            self._team_id = sly_env.team_id()

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        data = {}
        data["teamId"] = self._team_id
        if self._session_id is not None:
            data["model_connected"] = True
            if self._model_info is None:
                data["model_info"] = self._get_info()
            else:
                data["model_info"] = self._model_info
        elif self._session_id is None and self._model_info is not None:
            data["model_info"] = self._model_info
            data["model_connected"] = True
        else:
            data["model_info"] = None
            data["model_connected"] = False

        return data

    def get_json_state(self):
        state = {}
        state["sessionId"] = self._session_id
        return state

    def set_session_id(self, session_id: int):
        self._session_id = session_id
        self._model_info = self._get_info()
        self.update_data()
        self.update_state()
        DataJson().send_changes()
        StateJson().send_changes()

    def set_model_info(
        self,
        session_id: Optional[int] = None,
        model_info: Optional[dict] = None,
    ):
        if session_id is None and model_info is None:
            raise ValueError("Both session_id and model_info can't be None.")

        self._session_id = session_id
        self._model_info = model_info
        self.update_data()
        self.update_state()
        DataJson().send_changes()
        StateJson().send_changes()

    def _get_info(self):
        session = Session(self._api, self._session_id)
        return session.get_human_readable_info(self._replace_none_with)

    @property
    def session_id(self):
        return self._session_id



import inspect
import time
import uuid
from typing import BinaryIO, Dict, List, Tuple

import numpy as np
from pydantic import ValidationError

from supervisely._utils import find_value_by_keys, get_valid_kwargs
from supervisely.annotation.label import Geometry, Label
from supervisely.annotation.obj_class import ObjClass
from supervisely.api.api import Api
from supervisely.api.module_api import ApiField
from supervisely.api.video.video_figure_api import FigureInfo
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.helpers import deserialize_geometry
from supervisely.geometry.polygon import Polygon
from supervisely.imaging import image as sly_image
from supervisely.nn.inference.inference_request import InferenceRequest
from supervisely.nn.inference.tracking.base_tracking import BaseTracking
from supervisely.nn.inference.tracking.tracker_interface import (
    TrackerInterface,
    TrackerInterfaceV2,
)
from supervisely.nn.inference.uploader import Uploader
from supervisely.sly_logger import logger


class MaskTracking(BaseTracking):
    """Video tracking for mask/polygon geometries; uses model to propagate segmentation across frames."""

    def _deserialize_geometry(self, data: dict):
        geometry_type_str = data["type"]
        geometry_json = data["data"]
        return deserialize_geometry(geometry_type_str, geometry_json)

    def _inference(self, frames: List[np.ndarray], geometries: List[Geometry]):
        results = [[] for _ in range(len(frames) - 1)]
        for geometry in geometries:
            if not isinstance(geometry, Bitmap) and not isinstance(geometry, Polygon):
                raise TypeError(f"This app does not support {geometry.geometry_name()} tracking")

            # combine several binary masks into one multilabel mask
            i = 0
            label2id = {}

            original_geometry = geometry.clone()

            # convert polygon to bitmap
            if isinstance(geometry, Polygon):
                polygon_obj_class = ObjClass("polygon", Polygon)
                polygon_label = Label(geometry, polygon_obj_class)
                bitmap_obj_class = ObjClass("bitmap", Bitmap)
                bitmap_label = polygon_label.convert(bitmap_obj_class)[0]
                geometry = bitmap_label.geometry
            if i == 0:
                multilabel_mask = geometry.data.astype(int)
                multilabel_mask = np.zeros(frames[0].shape, dtype=np.uint8)
                geometry.draw(bitmap=multilabel_mask, color=[1, 1, 1])
                i += 1
            else:
                i += 1
                geometry.draw(bitmap=multilabel_mask, color=[i, i, i])
            label2id[i] = {
                "original_geometry": original_geometry.geometry_name(),
            }
            # run tracker
            tracked_multilabel_masks = self.predict(
                frames=frames, input_mask=multilabel_mask[:, :, 0]
            )
            tracked_multilabel_masks = np.array(tracked_multilabel_masks)
            # decompose multilabel masks into binary masks
            for i in np.unique(tracked_multilabel_masks):
                if i != 0:
                    binary_masks = tracked_multilabel_masks == i
                    geometry_type = label2id[i]["original_geometry"]
                    for j, mask in enumerate(binary_masks[1:]):
                        # check if mask is not empty
                        if np.any(mask):
                            if geometry_type == "polygon":
                                bitmap_geometry = Bitmap(mask)
                                bitmap_obj_class = ObjClass("bitmap", Bitmap)
                                bitmap_label = Label(bitmap_geometry, bitmap_obj_class)
                                polygon_obj_class = ObjClass("polygon", Polygon)
                                polygon_labels = bitmap_label.convert(polygon_obj_class)
                                geometries = [label.geometry for label in polygon_labels]
                            else:
                                geometries = [Bitmap(mask)]
                            results[j].extend(
                                [
                                    {"type": geom.geometry_name(), "data": geom.to_json()}
                                    for geom in geometries
                                ]
                            )
        return results

    def _track(self, api: Api, context: Dict, inference_request: InferenceRequest):
        video_interface = TrackerInterface(
            context=context,
            api=api,
            load_all_frames=False,
            notify_in_predict=True,
            per_point_polygon_tracking=False,
            frame_loader=self.cache.download_frame,
            frames_loader=self.cache.download_frames,
        )
        video_interface.stop += video_interface.frames_count + 1
        range_of_frames = [
            video_interface.frames_indexes[0],
            video_interface.frames_indexes[-1],
        ]

        if self.cache.is_persistent:
            # if cache is persistent, run cache task for whole video
            self.cache.run_cache_task_manually(
                api,
                None,
                video_id=video_interface.video_id,
            )
        else:
            # if cache is not persistent, run cache task for range of frames
            self.cache.run_cache_task_manually(
                api,
                [range_of_frames],
                video_id=video_interface.video_id,
            )

        api.logger.debug("frames_count = %s", video_interface.frames_count)
        inference_request.set_stage("Downloading frames", 0, video_interface.frames_count)
        # load frames

        def _load_frames_cb(n: int = 1):
            inference_request.done(n)
            video_interface._notify(pos_increment=n, task="Downloading frames")

        frames = self.cache.download_frames(
            api,
            video_interface.video_id,
            video_interface.frames_indexes,
            progress_cb=_load_frames_cb,
        )

        # combine several binary masks into one multilabel mask
        i = 1
        label2id = {}
        multilabel_mask = np.zeros(frames[0].shape, dtype=np.uint8)
        for (fig_id, geometry), obj_id in zip(
            video_interface.geometries.items(),
            video_interface.object_ids,
        ):
            original_geometry = geometry.clone()
            if not isinstance(geometry, Bitmap) and not isinstance(geometry, Polygon):
                raise TypeError(f"This app does not support {geometry.geometry_name()} tracking")
            # convert polygon to bitmap
            if isinstance(geometry, Polygon):
                polygon_obj_class = ObjClass("polygon", Polygon)
                polygon_label = Label(geometry, polygon_obj_class)
                bitmap_obj_class = ObjClass("bitmap", Bitmap)
                bitmap_label = polygon_label.convert(bitmap_obj_class)[0]
                geometry = bitmap_label.geometry
            geometry.draw(bitmap=multilabel_mask, color=i)
            label2id[i] = {
                "fig_id": fig_id,
                "obj_id": obj_id,
                "original_geometry": original_geometry.geometry_name(),
            }
            i += 1

        unique_labels = np.unique(multilabel_mask)
        if 0 in unique_labels:
            unique_labels = unique_labels[1:]
        api.logger.debug("unique_labels = %s", unique_labels)
        total_progress = len(unique_labels) * video_interface.frames_count
        api.logger.info("Starting tracking process")
        api.logger.debug("total_progress = %s", total_progress)
        inference_request.set_stage(
            InferenceRequest.Stage.INFERENCE,
            0,
            total_progress,
        )

        def _upload_f(items: List):
            video_interface.add_object_geometries_on_frames(*list(zip(*items)))
            inference_request.done(sum(item[-1] for item in items))

        with Uploader(upload_f=_upload_f, logger=api.logger) as uploader:
            # run tracker
            tracked_multilabel_masks = self.predict(
                frames=frames, input_mask=multilabel_mask[:, :, 0]
            )
            for curframe_i, mask in enumerate(
                tracked_multilabel_masks, video_interface.frame_index
            ):
                if curframe_i == video_interface.frame_index:
                    continue
                for i in unique_labels:
                    binary_mask = mask == i
                    fig_id = label2id[i]["fig_id"]
                    obj_id = label2id[i]["obj_id"]
                    geometry_type = label2id[i]["original_geometry"]
                    if not np.any(binary_mask):
                        api.logger.info(f"Skipping empty mask on frame {curframe_i}")
                        inference_request.done()
                    else:
                        if geometry_type == "polygon":
                            bitmap_geometry = Bitmap(binary_mask)
                            bitmap_obj_class = ObjClass("bitmap", Bitmap)
                            bitmap_label = Label(bitmap_geometry, bitmap_obj_class)
                            polygon_obj_class = ObjClass("polygon", Polygon)
                            polygon_labels = bitmap_label.convert(polygon_obj_class)
                            geometries = [label.geometry for label in polygon_labels]
                        else:
                            geometries = [Bitmap(binary_mask)]
                        uploader.put(
                            [
                                (
                                    geometry,
                                    obj_id,
                                    curframe_i,
                                    True if g_idx == len(geometries) - 1 else False,
                                )
                                for g_idx, geometry in enumerate(geometries)
                            ]
                        )
                    if inference_request.is_stopped() or video_interface.global_stop_indicatior:
                        api.logger.info(
                            "Tracking stopped by user",
                            extra={"inference_request_uuid": inference_request.uuid},
                        )
                        video_interface._notify(True, task="Stop tracking")
                        return
                    if uploader.has_exception():
                        raise uploader.exception

                api.logger.info(f"Frame {curframe_i} was successfully tracked")

    def _track_async(self, api: Api, context: dict, inference_request: InferenceRequest):
        tracker_interface = TrackerInterfaceV2(api, context, self.cache)
        frames_count = tracker_interface.frames_count
        figures = tracker_interface.figures
        progress_total = frames_count * len(figures)
        frame_range = [
            tracker_interface.frame_indexes[0],
            tracker_interface.frame_indexes[-1],
        ]
        frame_range_asc = [min(frame_range), max(frame_range)]

        def _upload_f(items: List[Tuple[FigureInfo, bool]]):
            inference_request.add_results([item[0] for item in items])
            inference_request.done(sum(item[1] for item in items))

        def _notify_f(items: List[Tuple[FigureInfo, bool]]):
            frame_range = [
                min(item[0].frame_index for item in items),
                max(item[0].frame_index for item in items),
            ]
            tracker_interface.notify_progress(
                inference_request.progress.current, inference_request.progress.total, frame_range
            )

        def _exception_handler(exception: Exception):
            api.logger.error(f"Error saving predictions: {str(exception)}", exc_info=True)
            tracker_interface.notify_progress(
                inference_request.progress.current,
                inference_request.progress.current,
                frame_range_asc,
            )
            tracker_interface.notify_error(exception)
            raise Exception

        def _maybe_stop():
            if inference_request.is_stopped() or tracker_interface.is_stopped():
                if isinstance(tracker_interface.stop_reason(), Exception):
                    raise tracker_interface.stop_reason()
                api.logger.info(
                    "Inference request stopped.",
                    extra={"inference_request_uuid": inference_request.uuid},
                )
                tracker_interface.notify_progress(
                    inference_request.progress.current,
                    inference_request.progress.current,
                    frame_range_asc,
                )
                return True
            if uploader.has_exception():
                raise uploader.exception
            return False

        # run tracker
        frame_index = tracker_interface.frame_index
        direction_n = tracker_interface.direction_n
        api.logger.info("Start tracking.")
        inference_request.set_stage(InferenceRequest.Stage.INFERENCE, 0, progress_total)
        with Uploader(
            upload_f=_upload_f,
            notify_f=_notify_f,
            exception_handler=_exception_handler,
            logger=api.logger,
        ) as uploader:
            # combine several binary masks into one multilabel mask
            i = 0
            label2id = {}
            # load frames
            frames = tracker_interface.load_all_frames()
            frames = [frame.image for frame in frames]
            for figure in figures:
                figure = api.video.figure._convert_json_info(figure)
                fig_id = figure.id
                obj_id = figure.object_id
                geometry = deserialize_geometry(figure.geometry_type, figure.geometry)
                original_geometry = geometry.clone()
                if not isinstance(geometry, (Bitmap, Polygon)):
                    raise TypeError(
                        f"This app does not support {geometry.geometry_name()} tracking"
                    )
                # convert polygon to bitmap
                if isinstance(geometry, Polygon):
                    polygon_obj_class = ObjClass("polygon", Polygon)
                    polygon_label = Label(geometry, polygon_obj_class)
                    bitmap_obj_class = ObjClass("bitmap", Bitmap)
                    bitmap_label = polygon_label.convert(bitmap_obj_class)[0]
                    geometry = bitmap_label.geometry
                if i == 0:
                    multilabel_mask = geometry.data.astype(int)
                    multilabel_mask = np.zeros(frames[0].shape, dtype=np.uint8)
                    geometry.draw(bitmap=multilabel_mask, color=[1, 1, 1])
                    i += 1
                else:
                    i += 1
                    geometry.draw(bitmap=multilabel_mask, color=[i, i, i])
                label2id[i] = {
                    "fig_id": fig_id,
                    "obj_id": obj_id,
                    "original_geometry": original_geometry.geometry_name(),
                }

                if _maybe_stop():
                    return

            # predict
            tracked_multilabel_masks = self.predict(
                frames=frames, input_mask=multilabel_mask[:, :, 0]
            )
            tracked_multilabel_masks = np.array(tracked_multilabel_masks)

            # decompose multilabel masks into binary masks
            for i in np.unique(tracked_multilabel_masks):
                if _maybe_stop():
                    return
                if i != 0:
                    binary_masks = tracked_multilabel_masks == i
                    fig_id = label2id[i]["fig_id"]
                    obj_id = label2id[i]["obj_id"]
                    geometry_type = label2id[i]["original_geometry"]
                    for j, mask in enumerate(binary_masks[1:], 1):
                        if _maybe_stop():
                            return
                        this_figure_index = frame_index + j * direction_n
                        # check if mask is not empty
                        if not np.any(mask):
                            api.logger.info(f"Skipping empty mask on frame {this_figure_index}")
                            # update progress bar anyway (otherwise it will not be finished)
                            inference_request.done()
                        else:
                            if geometry_type == "polygon":
                                bitmap_geometry = Bitmap(mask)
                                bitmap_obj_class = ObjClass("bitmap", Bitmap)
                                bitmap_label = Label(bitmap_geometry, bitmap_obj_class)
                                polygon_obj_class = ObjClass("polygon", Polygon)
                                polygon_labels = bitmap_label.convert(polygon_obj_class)
                                geometries = [label.geometry for label in polygon_labels]
                            else:
                                geometries = [Bitmap(mask)]
                            for l, geometry in enumerate(geometries):
                                figure_id = uuid.uuid5(
                                    namespace=uuid.NAMESPACE_URL, name=f"{time.time()}"
                                ).hex
                                result_figure = api.video.figure._convert_json_info(
                                    {
                                        ApiField.ID: figure_id,
                                        ApiField.OBJECT_ID: obj_id,
                                        "meta": {"frame": this_figure_index},
                                        ApiField.GEOMETRY_TYPE: geometry.geometry_name(),
                                        ApiField.GEOMETRY: geometry.to_json(),
                                        ApiField.TRACK_ID: tracker_interface.track_id,
                                    }
                                )
                                should_notify = l == len(geometries) - 1
                                tracker_interface.add_prediction((result_figure, should_notify))
                    api.logger.info(
                        "Figure [%d, %d] tracked.",
                        i,
                        len(figures),
                        extra={"figure_id": figure.id},
                    )

    def _track_api(self, api: Api, context: Dict, inference_request: InferenceRequest):
        # unused fields:
        context["trackId"] = "auto"
        context["objectIds"] = []
        context["figureIds"] = []
        if "direction" not in context:
            context["direction"] = "forward"

        input_geometries: list = context["input_geometries"]

        video_interface = TrackerInterface(
            context=context,
            api=api,
            load_all_frames=True,
            notify_in_predict=True,
            per_point_polygon_tracking=False,
            frame_loader=self.cache.download_frame,
            frames_loader=self.cache.download_frames,
            should_notify=False,
        )

        range_of_frames = [
            video_interface.frames_indexes[0],
            video_interface.frames_indexes[-1],
        ]

        if self.cache.is_persistent:
            # if cache is persistent, run cache task for whole video
            self.cache.run_cache_task_manually(
                api,
                None,
                video_id=video_interface.video_id,
            )
        else:
            # if cache is not persistent, run cache task for range of frames
            self.cache.run_cache_task_manually(
                api,
                [range_of_frames],
                video_id=video_interface.video_id,
            )

        inference_request.set_stage("Downloading frames", 0, video_interface.frames_count)
        # load frames
        frames = self.cache.download_frames(
            api,
            video_interface.video_id,
            video_interface.frames_indexes,
            progress_cb=inference_request.done,
        )
        # combine several binary masks into one multilabel mask
        label2id = {}

        multilabel_mask = np.zeros(frames[0].shape, dtype=np.uint8)
        for i, input_geom in enumerate(input_geometries, 1):
            geometry = self._deserialize_geometry(input_geom)
            if not isinstance(geometry, Bitmap) and not isinstance(geometry, Polygon):
                raise TypeError(f"This app does not support {geometry.geometry_name()} tracking")
            # convert polygon to bitmap
            if isinstance(geometry, Polygon):
                polygon_obj_class = ObjClass("polygon", Polygon)
                polygon_label = Label(geometry, polygon_obj_class)
                bitmap_obj_class = ObjClass("bitmap", Bitmap)
                bitmap_label = polygon_label.convert(bitmap_obj_class)[0]
                geometry = bitmap_label.geometry
            geometry.draw(bitmap=multilabel_mask, color=i)
            label2id[i] = {
                "original_geometry": geometry.geometry_name(),
            }

        result_indexes = np.unique(multilabel_mask)
        progress_total = len(result_indexes)
        if 0 in result_indexes:
            progress_total -= 1
        progress_total = progress_total * video_interface.frames_count

        api.logger.info("Starting tracking process")
        inference_request.set_stage(
            InferenceRequest.Stage.INFERENCE,
            0,
            progress_total,
        )

        # run tracker
        tracked_multilabel_masks = self.predict(frames=frames, input_mask=multilabel_mask[:, :, 0])
        tracked_multilabel_masks = np.array(tracked_multilabel_masks)

        predictions = []
        # decompose multilabel masks into binary masks
        for i in np.unique(tracked_multilabel_masks):
            if i != 0:
                predictions_for_label = []
                binary_masks = tracked_multilabel_masks == i
                geometry_type = label2id[i]["original_geometry"]
                for j, mask in enumerate(binary_masks[1:]):
                    # check if mask is not empty
                    if not np.any(mask):
                        api.logger.info(f"Empty mask on frame {context['frameIndex'] + j + 1}")
                        predictions_for_label.append(None)
                    else:
                        # frame_idx = j + 1
                        geometry = Bitmap(mask)
                        predictions_for_label.append(
                            {"type": geometry.geometry_name(), "data": geometry.to_json()}
                        )
                predictions.append(predictions_for_label)

        # predictions must be NxK masks: N=number of frames, K=number of objects
        predictions = list(map(list, zip(*predictions)))
        inference_request.final_result = predictions
        return predictions

    # Implement the following methods in the derived class
    def track(self, api: Api, state: Dict, context: Dict):
        fn = self.send_error_data(api, context)(self._track)
        self.inference_requests_manager.schedule_task(fn, api, context)
        return {"message": "Tracking has started."}

    def track_api(self, api: Api, state: Dict, context: Dict):
        inference_request, future = self.inference_requests_manager.schedule_task(
            self._track_api, api, context
        )
        future.result()
        logger.info(
            "Track-api request processed.", extra={"inference_request_uuid": inference_request.uuid}
        )
        return inference_request.final_result

    def track_api_files(
        self,
        files: List[BinaryIO],
        settings: Dict,
    ):
        logger.info(f"Start tracking with settings:", extra={"settings": settings})
        frame_index = find_value_by_keys(settings, ["frame_index", "frameIndex"])
        frames_count = find_value_by_keys(settings, ["frames", "framesCount"])
        input_geometries = find_value_by_keys(settings, ["input_geometries", "inputGeometries"])
        direction = settings.get("direction", "forward")
        direction_n = 1 if direction == "forward" else -1
        frame_indexes = list(
            range(frame_index, frame_index + frames_count * direction_n + direction_n, direction_n)
        )
        geometries = map(self._deserialize_geometry, input_geometries)
        frames = []
        for file, frame_idx in zip(files, frame_indexes):
            img_bytes = file.read()
            frame = sly_image.read_bytes(img_bytes)
            frames.append(frame)
        logger.info("Start tracking.")
        return self._inference(frames, geometries)

    def track_async(self, api: Api, state: Dict, context: Dict):
        batch_size = context.get("batch_size", self.get_batch_size())
        if self.max_batch_size is not None and batch_size > self.max_batch_size:
            raise ValidationError(
                f"Batch size should be less than or equal to {self.max_batch_size} for this model."
            )

        fn = self.send_error_data(api, context)(self._track_async)
        inference_request, _ = self.inference_requests_manager.schedule_task(fn, api, context)

        logger.debug(
            "Inference has scheduled from 'track_async' endpoint",
            extra={"inference_request_uuid": inference_request.uuid},
        )
        return {
            "message": "Inference has started.",
            "inference_request_uuid": inference_request.uuid,
        }
